/* $Id: cntrl.c,v 1.29 2007-01-22 21:38:18 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <err.h>
#include <limits.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#include "pl_mpi.h"

/*-[ type declarations ]---------------------------------------------*/

/*-[ function declarations ]-----------------------------------------*/

/** Process a "start" request from the master node.  Initialises a job
 * and does the initial fan-out to accepting nodes.
 * @param source The master node's rank (always 0). */
static void
do_cntrl_firstcall(int source);

/** Process a "redo" request from the master node.  Returns available
 * results or marks as "blocking" for later return. 
 * @param source The master node's rank (always 0). */
static void
do_cntrl_redo(int source);

static void
do_cntrl_fanout(struct cntrl_pending *pend);

/** Process a "cut" request from the master node.
 * @param source The master node's rank (always 0). */
static void
do_cntrl_cut(int source);

/* Process a slave's notification that it has finished its job.
 * @param source The slave finishing its job. */
static void
do_cntrl_done(int source);

/*-[ variable definitions ]------------------------------------------*/

static struct cntrl_pending **workers;

/* A queue buffer of pending jobs. */
#define RUN_QUEUE_SZ	256
static struct cntrl_pending *run_queue[RUN_QUEUE_SZ];
static unsigned int run_queue_size = 0;

static unsigned long id_current = 0;

/*-[ function definitions ]------------------------------------------*/

/** Free up a cntrl_pending structure's memory. */
static void
job_free(struct cntrl_pending *pend)
{
	unsigned int i;
	assert(pend != NULL);
	assert(pend->pred != NULL);
	assert(pend->data != NULL);
	for (i = 0; i < pend->count; i++) {
		if (pend->data[i].input != NULL)
			free(pend->data[i].input);
		if (pend->data[i].output != NULL)
			free(pend->data[i].output);
	}
	free(pend->pred);
	free(pend->data);
	free(pend);
}

static struct cntrl_pending *
job_init(void)
{
	struct cntrl_pending *p;
	p = calloc(1, sizeof(struct cntrl_pending));
	assert(p != NULL);
	p->ident = id_current++;
	return(p);
}

/** Add a cntl_pending structure to the end of the queue buffer. */
static void
job_enqueue(struct cntrl_pending *pend)
{
	int i;
	assert(pend != NULL);
	for (i = 0; run_queue[i] != NULL && i < RUN_QUEUE_SZ; i++) 
		/* Do nothing. */ ;
	assert(i < RUN_QUEUE_SZ);
	assert(run_queue[i] == NULL);
	run_queue[i] = pend;
	run_queue_size++;
}

/** Remove a cntrl_pending structure from the queue buffer and adjust the
 * remaining elements. */
static void
job_dequeue(struct cntrl_pending *pend)
{
	int i;
	assert(pend != NULL);
	for (i = 0; run_queue[i] != NULL && i < RUN_QUEUE_SZ; i++)
		if (run_queue[i] == pend)
			break;
	assert(i < RUN_QUEUE_SZ);
	assert(run_queue[i] == pend);
	run_queue[i] = NULL;
	/* Shift the rest to the left. */
	i++;
	for ( ; run_queue[i] != NULL && i < RUN_QUEUE_SZ; i++) {
		run_queue[i - 1] = run_queue[i];
		run_queue[i] = NULL;
	}
	run_queue_size--;
}

/*static*/ void
do_cntrl_fanout(struct cntrl_pending *pend)
{
	unsigned int i, j, k, l;
		/* i <- node rank index */
		/* j <- argument index in per-job iteration */
		/* k <- index (modulo run_queue_size) in job queue */
		/* l <- number of scanned jobs for new slots */
	char code;
	struct cntrl_pending *job;
	MPI_Request rq;
	assert(workers != NULL);

	k = 0;
	/* k = index in run_queue of current entry */
	if (pend != NULL) for ( ; k < run_queue_size && 
			run_queue[k] != pend; k++)
		/* Do nothing. */ ;

	/* Be very careful in modifying sections of this code.  This
	 * governs the job scheduler.  At this time, the scheduler uses
	 * a "fair-sharing" algorithm that equally serves all queued
	 * jobs as workers free up.  */

	code = TAG_EVENT_WORKER;
	for (i = 2; i < nodes; i++) {
		/* See if a node is still processing. */
		if (workers[i] != NULL)
			continue;

		/* Find the job to consider. */
		for (l = 0; l < run_queue_size; l++, k++) {
			job = run_queue[k % run_queue_size];
			assert(job != NULL);
			for (j = 0; j < job->count; j++)
				if (job->data[j].status == STATUS_NEW)
					break;
			if (j < job->count)
				break;
			MPI_DEBUG("job %p (%lu) (%u) exhausted (%u)", 
					job, job->ident, 
					k % run_queue_size, j);
		}
		/* Check if we've no jobs left to distribute. */
		if (l >= run_queue_size) {
			MPI_DEBUG("job queue exhausted (%u)", l);
			return;
		}
		MPI_DEBUG("job %p (%lu) (%u) slot %u of %u to %u", 
				job, job->ident, k % run_queue_size, 
				j, job->count, i);

		if (gettimeofday(&job->data[j].start, NULL) == -1)
			MPI_ABORT();

		/* Send out to node. */
		MPI_ISEND(&code, 1, MPI_CHAR, i, TAG_EVENT, &rq);
		MPI_ISEND(&job->ident, 1, MPI_UNSIGNED_LONG, i,
				TAG_WORKER_PREDKEY, &rq);
		MPI_ISEND(&j, 1, MPI_UNSIGNED, i,
				TAG_WORKER_PREDINDEX, &rq);
		MPI_ISEND(&job->predlen, 1, MPI_UNSIGNED, i, 
				TAG_WORKER_PREDSIZE, &rq);
		MPI_ISEND(job->pred, job->predlen, MPI_CHAR,
				i, TAG_WORKER_PRED, &rq);
		/* Send out request-specific stuff. */
		MPI_ISEND(&job->data[j].input_type, 1, MPI_UNSIGNED,
				i, TAG_WORKER_DATATYPE, &rq);
		MPI_ISEND(&job->data[j].input_length, 1, MPI_UNSIGNED,
				i, TAG_WORKER_DATASIZE, &rq);
		MPI_ISEND(job->data[j].input, job->data[j].input_length, 
				MPI_CHAR, i, TAG_WORKER_DATA, &rq);

		/* Update node and worker status. */
		workers[i] = job;
		job->data[j].node = i;
		job->data[j].status = STATUS_WAITING;

		MPI_TRACE("%10s %10lu %10u", "start", 
				job->ident, j);
	}
}

void
do_cntrl_cut(int source)
{
	struct cntrl_pending *pend;
	unsigned int i;

	MPI_RECV(&pend, (int)sizeof(void *), MPI_CHAR,
			source, TAG_CUT_HANDLE);
	assert( ! (pend->flags & FLAGS_CUTTED));
	assert( ! (pend->flags & FLAGS_BLOCK));
	/* Mark as cutted. */
	pend->flags |= FLAGS_CUTTED;

	MPI_TRACE("%10s %10lu", "cut", pend->ident);

	/* Scan through and see if we've a pending job; exit of so. */
	for (i = 0; i < nodes; i++)
		if (workers[i] == pend) {
			MPI_DEBUG("job %p (%lu) queued for release", 
					pend, pend->ident);
			return;
		}

	MPI_DEBUG("job %p (%lu) cut and released", pend, pend->ident);

	/* No nodes are processing this job. Delete it. */
	MPI_TRACE("%10s %10lu", "free", pend->ident);
	job_free(pend);
	job_dequeue(pend);
}

void
do_cntrl_done(int source)
{
	struct cntrl_pending *pend;
	struct params *parms;
	unsigned int i;
	unsigned long long tt;
	struct timeval tv;
	MPI_Request rq;

	assert(workers != NULL);
	pend = workers[source];
	assert(pend != NULL);

	/* Sets "i" to be our relevant argument node. */
	for (i = 0; i < pend->count; i++) 
		if (pend->data[i].status == STATUS_WAITING &&
				pend->data[i].node == 
				(unsigned int)source)
			break;
	assert(i < pend->count);
	parms = &pend->data[i];

	MPI_RECV(&parms->output_yesno, 1, MPI_INT, source,
			TAG_COMPLETE_YESNO);

	MPI_DEBUG("trigger %p (%lu) slot %u from %d",
			pend, pend->ident, i, source);

	/* Get the results, if we have them. */
	if (parms->output_yesno) {
		MPI_RECV(&parms->output_type, 1, MPI_INT,
				source, TAG_COMPLETE_RESULTTYPE);
		MPI_RECV(&parms->output_length, 1, MPI_UNSIGNED,
				source, TAG_COMPLETE_RESULTSIZE);
		parms->output = calloc(parms->output_length + 1, 1);
		assert(parms->output != NULL);
		MPI_RECV(parms->output, parms->output_length, 
				MPI_CHAR, source, TAG_COMPLETE_RESULT);
		MPI_DEBUG("affirm %p (%lu) slot %u (type %d, "
				"size %d) from %d",
				pend, pend->ident, i,
				parms->output_type, 
				parms->output_length, source);
	} else {
		MPI_DEBUG("failure %p (%lu) slot %u from %d",
				pend, pend->ident, i, source);
	}

	if (gettimeofday(&parms->end, NULL) == -1)
		MPI_ABORT();

	MPI_TRACE("%10s %10lu %10u", "complete", 
			pend->ident, i);

	timersub(&parms->end, &parms->start, &tv);
	/* This can be less than zero on some systems. */
	if (tv.tv_sec < 0)
		tv.tv_sec = tv.tv_usec = 0;
	tv.tv_usec = tv.tv_usec < 0 ? 0 : tv.tv_usec;
	/* Cast aggr. to larger-width type. */
	tt = ((unsigned long long)tv.tv_sec * 1000000) +
		(unsigned long long)tv.tv_usec;
	if (opts.dump != NULL) {
		if (fprintf(opts.dump, 
				"%10lu %7u %7u %20llu\n", 
				pend->ident, i, source, 
				tt) == -1)
			MPI_ABORT();
		if (fflush(opts.dump) == -1)
			MPI_ABORT();
	}

	/* Mark ourselves as being complete. */
	workers[source] = NULL;
	parms->status = STATUS_COMPLETE;

	/* Send to the root, if it's blocking on results. */
	if (pend->flags & FLAGS_BLOCK) {
		assert( ! (pend->flags & FLAGS_CUTTED));
		parms->status = STATUS_CONFIRMED;
		/* NB: this is the argument, not host, index. */
		MPI_ISEND((int *)&i, 1, MPI_INT, RANK_ROOT, 
				TAG_REDO_INDEX, &rq);
		MPI_ISEND(&parms->output_yesno, 1, MPI_INT, RANK_ROOT,
				TAG_REDO_RESULTYES, &rq);
		if (parms->output_yesno) {
			MPI_ISEND(&parms->output_type, 1, MPI_INT, 
					RANK_ROOT, TAG_REDO_RESULTTYPE, 
					&rq);
			MPI_ISEND(&parms->output_length, 1, 
					MPI_UNSIGNED, RANK_ROOT, 
					TAG_REDO_RESULTSIZE, &rq);
			MPI_ISEND(parms->output, parms->output_length, 
					MPI_CHAR, RANK_ROOT, 
					TAG_REDO_RESULT, &rq);
		}
		pend->flags &= ~FLAGS_BLOCK;
		MPI_DEBUG("job %p (%lu) done (%d), un-block master", 
				pend, pend->ident, source);
	}

	/* Conditionally clean up this predicate, if it's no longer a
	 * valid one. */
	if (pend->flags & FLAGS_CUTTED) {
		/* Scan for nodes containing the same. */
		for (i = 0; i < nodes; i++)
			if (workers[i] == pend)
				break;
		/* No nodes processing our job.  Delete it. */
		if (i >= nodes) {
			MPI_TRACE("%10s %10lu", "free", pend->ident);
			job_free(pend);
			job_dequeue(pend);
			/* No need to stick around. */
			MPI_DEBUG("job %p (%lu) done (%d), slot clear", 
					pend, pend->ident, source);
			return;
		}
	}
	do_cntrl_fanout(pend);
}

void
do_cntrl_redo(int source)
{
	struct cntrl_pending *pend;
	struct params *parms;
	unsigned int nconf, i;
	int res;
	MPI_Request rq;

	MPI_RECV(&pend, (int)sizeof(struct cntrl_pending *),
			MPI_CHAR, source, TAG_REDO_HANDLE);

	assert( ! (pend->flags & FLAGS_BLOCK));

	/* First scan to see if we've no results.  This occurs
	 * out-of-line to keep the code cleaner. */
	for (i = 0, nconf = 0; i < pend->count; i++)
		if (pend->data[i].status == STATUS_CONFIRMED)
			nconf++;
		else if (pend->data[i].status == STATUS_COMPLETE)
			break;
	/* Are we the last entry? */
	if (i >= pend->count) {
		if (nconf < pend->count)  {
			pend->flags |= FLAGS_BLOCK;
			return;
		}
		res = RESULT_FAILURE;
		MPI_ISEND(&res, 1, MPI_INT, RANK_ROOT, 
				TAG_REDO_INDEX, &rq);
		return;
	}
	assert(i < INT_MAX);
	res = (int)i;
	parms = &pend->data[i];

	parms->status = STATUS_CONFIRMED;

	MPI_ISEND(&res, 1, MPI_INT, RANK_ROOT, TAG_REDO_INDEX, &rq);
	MPI_ISEND(&parms->output_yesno, 1, MPI_INT, RANK_ROOT, 
			TAG_REDO_RESULTYES, &rq);
	if (parms->output_yesno) {
		MPI_ISEND(&parms->output_type, 1, MPI_INT, 
				RANK_ROOT, TAG_REDO_RESULTTYPE, &rq);
		MPI_ISEND(&parms->output_length, 1, MPI_UNSIGNED, 
				RANK_ROOT, TAG_REDO_RESULTSIZE, &rq);
		MPI_ISEND(parms->output, parms->output_length, 
				MPI_CHAR, RANK_ROOT, TAG_REDO_RESULT, 
				&rq);
	}
}

void
do_cntrl_query(int source)
{
	struct cntrl_pending *pend;
	MPI_Request rq;
	unsigned int i;
	long count;

	MPI_RECV(&pend, (int)sizeof(struct cntrl_pending *),
			MPI_CHAR, source, TAG_QUERY_HANDLE);
	MPI_TRACE("%10s %10lu", "query", pend->ident);
	for (count = 0, i = 0; i < pend->count; i++)
		if (pend->data[i].status == STATUS_COMPLETE)
			count++;
	MPI_ISEND(&count, 1, MPI_LONG, source, 
			TAG_QUERY_RESULT, &rq);
}

void
do_cntrl_firstcall(int source)
{
	unsigned int i, size;
	struct cntrl_pending *pend;
	struct params *parms;
	MPI_Request rq;

	/* Create & initialise "pending" structure. */
	pend = job_init();
	MPI_TRACE("%10s %10lu", "init", pend->ident);

	/* Read the predicate name & cache its length. */
	MPI_RECV(&size, 1, MPI_UNSIGNED, source, 
			TAG_FIRSTCALL_PREDSIZE);
	pend->pred = calloc(size + 1, 1);
	assert(pend->pred != NULL);
	MPI_RECV(pend->pred, size, MPI_CHAR, source, 
			TAG_FIRSTCALL_PRED);
	pend->predlen = size;

	/* Get number of input items. */
	MPI_RECV(&pend->count, 1, MPI_UNSIGNED, source, 
			TAG_FIRSTCALL_COUNT);

	pend->data = calloc(pend->count, sizeof(struct params));
	assert(pend->data != NULL);

	/* Actually read in strings to structure. */
	for (i = 0; i < pend->count; i++) {
		parms = &pend->data[i];
		MPI_RECV(&parms->input_type, 1, MPI_UNSIGNED,
				source, TAG_FIRSTCALL_DATATYPE);
		MPI_RECV(&parms->input_length, 1, MPI_UNSIGNED, 
				source, TAG_FIRSTCALL_DATASIZE);
		parms->input = calloc(parms->input_length + 1, 1);
		assert(parms->input != NULL);
		MPI_RECV(parms->input, parms->input_length,
				MPI_CHAR, source, TAG_FIRSTCALL_DATA);
	}

	MPI_DEBUG("job %p (%lu) beginning (%s)", pend, 
			pend->ident, pend->pred);
	job_enqueue(pend);

	/* Return the key for this entry. */
	MPI_ISEND(&pend, (int)sizeof(void *), MPI_CHAR,
			RANK_ROOT, TAG_FIRSTCALL_HANDLE, &rq);
	MPI_ISEND(&pend->ident, 1, MPI_UNSIGNED_LONG,
			RANK_ROOT, TAG_FIRSTCALL_IDENT, &rq);

	do_cntrl_fanout(NULL);
}

/*extern*/
int
loop_cntrl(void)
{
	int rc;
	char code;
	MPI_Status st;

	/* Allocate worker references. */
	workers = (struct cntrl_pending **)calloc
		(nodes, sizeof(struct cntrl_pending *));
	assert(workers != NULL);

	(void)memset(run_queue, 0, sizeof(struct cntrl_pending *) * 
			RUN_QUEUE_SZ);

	for ( ; ; ) { /* Infinite loop. */
		MPI_RECV_ST(&code, 1, MPI_CHAR, MPI_ANY_SOURCE, 
				TAG_EVENT, &st);
		switch(code) {
		case(TAG_EVENT_HALT):
			/* Free all remaining jobs? */
			free(workers);
			goto out;
		case(TAG_EVENT_FIRSTCALL):
			do_cntrl_firstcall(st.MPI_SOURCE);
			break;
		case(TAG_EVENT_REDO):
			do_cntrl_redo(st.MPI_SOURCE);
			break;
		case(TAG_EVENT_COMPLETE):
			do_cntrl_done(st.MPI_SOURCE);
			break;
		case(TAG_EVENT_CUT):
			do_cntrl_cut(st.MPI_SOURCE);
			break;
		case(TAG_EVENT_QUERY):
			do_cntrl_query(st.MPI_SOURCE);
			break;
		default:
			MPI_ABORT();
		}
	} /* Infinite loop. */

out: /* Egress point. */
	MPI_RECV(&rc, 1, MPI_INT, RANK_ROOT, TAG_HALT_CODE);
	MPI_BARRIER();
	MPI_FINALIZE();

	return(rc);
}

