/* $Id: dtime.c,v 1.3 2007-01-24 12:11:52 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/ioctl.h>
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "profile.h"

#ifndef INTERVAL
# define INTERVAL 		0
#endif /*!INTERVAL*/

#define US2MS(val) ((long double)(val) / 1000)
#define US2S(val) ((long double)(val) / 1000000)

/*-[variable definitions]--------------------------------------------*/

#define DEFAULT_TERMWIDTH	80
#define DEFAULT_TERMHEIGHT	45
static unsigned short termwidth = DEFAULT_TERMWIDTH;
static unsigned short termheight = DEFAULT_TERMHEIGHT;

static int show_breakdown = 0;
static int daemon_mode = 0;

/*-[function definitions]--------------------------------------------*/

/** Query for the terminal's dimensions. */
static void
dowinsize(void)
{
	struct winsize ws;
	if (ioctl(1, TIOCGWINSZ, (void *)&ws) != 0) {
		termwidth = DEFAULT_TERMWIDTH;
		termheight = DEFAULT_TERMHEIGHT;
	}
	termwidth = ws.ws_col;
	termheight = ws.ws_row;
}

/** Handle a SIGWINCH and update our dimensions. */
/*anon*/ void
dosigwinch(int sig)
{
	assert(sig == SIGWINCH);
	dowinsize();
}

/** Comparison function passed along to qsort. */
/*anon*/ int
qsort_cmp(const void *p1, const void *p2)
{
	return(*(const unsigned long long *)p1 -
			*(const unsigned long long *)p2);
}

/** @return TRUE on success, FALSE on failure. */
static int
doident(const struct master *master, const struct io_source *src,
		size_t size, const struct io_source *cntrl)
{
	unsigned long long cum_cntrl, med_cntrl,
		      cum_slave, med_slave, ttime,
		      *all_cntrl, *all_slave, *copy;
		/* all_cntrl and all_slave keyed by index */
	unsigned int *all_node, args1, args2;
		/* all_node keyed by index */
	unsigned long id;
	size_t i, j;
	assert(src != NULL);
	assert(master != NULL);
	assert(cntrl != NULL);

	id = master->id;
	if ((ttime = master->wallc) == 0)
		ttime = 1;

	/* Determine the number of control per-index entries. */
	for (args1 = 0, i = 0; i < cntrl->elems; i++)
		if (cntrl->data.cntrl[i].id == id)
			args1++;
	/* Determine the number of slave per-index entries. */
	for (args2 = 0, i = 0; i < size; i++)
		if (src[i].type == IO_SOURCE_SLAVE)
			for (j = 0; j < src[i].elems; j++) 
				if (src[i].data.slave[j].id == id) 
					args2++;
	if (args1 != args2)
		return(1);
	if (args1 == 0)
		return(1);

	cum_cntrl = cum_slave = 0;
	med_cntrl = med_slave = 0;

	/* Construct the control array & cumulative value. */
	all_cntrl = calloc(args1, sizeof(unsigned long long));
	assert(all_cntrl != NULL);
	/* Also construct the node-per-index array. */
	all_node = calloc(args1, sizeof(unsigned int));
	assert(all_node != NULL);
	for (i = 0; i < cntrl->elems; i++)
		if (cntrl->data.cntrl[i].id == id) {
			all_node[cntrl->data.cntrl[i].index] =
				cntrl->data.cntrl[i].slave;
			all_cntrl[cntrl->data.cntrl[i].index] =
				cntrl->data.cntrl[i].wallc == 0 ?
				1 : cntrl->data.cntrl[i].wallc;
			cum_cntrl += all_cntrl
				[cntrl->data.cntrl[i].index];
		}

	/* Construct the slave array & cumulative value. */
	all_slave = calloc(args1, sizeof(unsigned long long));
	assert(all_slave != NULL);
	for (i = 0; i < size; i++)
		if (src[i].type == IO_SOURCE_SLAVE)
			for (j = 0; j < src[i].elems; j++) {
				if (src[i].data.slave[j].id != id) 
					continue;
				/* Also normalise data. */
				all_slave[src[i].data.slave[j].index]
					= src[i].data.slave[j].wallc
					== 0 ? 1 :
					src[i].data.slave[j].wallc;
				cum_slave += all_slave
					[src[i].data.slave[j].index];
			}

	/* Construct our copy-array for median sort. */
	copy = calloc(args1, sizeof(unsigned long long));
	assert(copy != NULL);
	(void)memcpy(copy, all_slave, 
			sizeof(unsigned long long) * args1);
	qsort(copy, args1, sizeof(unsigned long long), qsort_cmp);
	/* (assign median value) */
	med_slave = copy[args1 / 2];
	(void)memcpy(copy, all_cntrl, 
			sizeof(unsigned long long) * args1);
	qsort(copy, args1, sizeof(unsigned long long), qsort_cmp);
	/* (assign median value) */
	med_cntrl = copy[args1 / 2];

	free(copy);

	cum_slave /= args2;
	cum_cntrl /= args1;

	/* Draw on screen. */

	(void)printf("%7lu | %10p |%15.4Lf |%10.2Lf %10.2Lf |"
			"%10.2Lf %10.2Lf | %7.2Lf%% %7.2Lf%%\n", 
			master->id, master->ctx,
			US2MS(ttime), 
			US2MS(cum_cntrl), US2MS(med_cntrl), 
			US2MS(cum_slave), US2MS(med_slave),
			((long double)cum_slave / cum_cntrl) * 100,
			((long double)med_slave / med_cntrl) * 100);

	if (show_breakdown) for (i = 0; i < args1; i++) 
		(void)printf("%32zd %4d |%21.4Lf |%21.4Lf | "
				"%16.2Lf%%\n",
				i, all_node[i],
				US2MS(all_cntrl[i]),
				US2MS(all_slave[i]),
				((long double)all_slave[i] / 
				 	all_cntrl[i]) * 100);
	free(all_cntrl);
	free(all_slave);
	free(all_node);
	return(1);
}

/*anon*/ int
doupdate(const struct io_source *src, size_t size)
{
	const struct io_source *master, *cntrl;
	size_t i;
	time_t t;
	assert(src != NULL);
	assert(size >= 3);

	/* First find the master and control nodes. */
	master = cntrl = NULL;
	for (i = 0; i < size; i++)
		if (src[i].type == IO_SOURCE_MASTER) 
			master = &src[i];
		else if (src[i].type == IO_SOURCE_CNTRL)
			cntrl = &src[i];
	assert(master != NULL);
	assert(cntrl != NULL);

	if (daemon_mode) {
		/* Upper-left, then clear. */
		(void)printf("\033[H\033[J");
		(void)fflush(stdout);
	}

	t = time(NULL);
	(void)printf("Query: %s", ctime(&t));
	(void)printf("%7s | %10s |%15s |%21s |%21s |%18s\n",
			"jid", "qid", "ttime", "cntrl avg/median",
			"slave avg/median", "% avg/median");

	if (show_breakdown) 
		(void)printf("%32s %4s |%21s |%21s |%18s\n",
				"index", "node", "cntrl",
				"slave", "%");

	for (i = 0; i < (size_t)termwidth; i++)
		(void)printf("-");
	(void)printf("\n");

	i = 0;
	if (daemon_mode && master->elems > 
			(unsigned long)(termheight - 3)) {
		i = master->elems - termheight + 5;
		(void)printf("(omitting %zd initial results)\n", i);
	}

	/* Iterate through the completed jobs. */
	for ( ; i < master->elems; i++)
		if ( ! doident(&master->data.master[i], 
					src, size, cntrl))
			return(-1);
	return(0);
}

static void
usage(int code)
{
	extern char *__progname;
	(void)fprintf(stderr, "Usage: %s [-b] [-i interval]"
			" file [...]\n", __progname);
	exit(code);
	/* NOTREACHED */
}

int
main(int argc, char *argv[])
{
	int c;
	const char *errstr;
	struct io_source src[MAX_SOURCES + 2];
	unsigned int interval;
	
	interval = INTERVAL;
	(void)memset(src, 0, sizeof(src));

	if (signal(SIGWINCH, dosigwinch) == SIG_ERR)
		err(1, "signal");
	dowinsize();

	while ((c = getopt(argc, argv, "bi:")) != -1) 
		switch (c) {
		case('i'):
			interval = (unsigned int)strtonum
				(optarg, 0, UINT_MAX, &errstr);
			if (errstr)
				errx(1, "interval is %s: %s",
						errstr, optarg);
			break;
		case('b'):
			show_breakdown = 1;
			break;
		default:
			usage(1);
			/* NOTREACHED */
		}
	argc -= optind;
	argv += optind;

	if (interval > 0)
		daemon_mode = 1;

	if ( ! pl_mpi_prof_init(src, MAX_SOURCES + 2,
				doupdate, argc, argv))
		return(1);
	return(pl_mpi_prof_run(src, argc, interval) ? 0 : 1);
}

