/* $Id: gdtime.c,v 1.1 2007-01-22 21:38:18 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/ioctl.h>
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "profile.h"

#ifndef GNUPLOT
# define GNUPLOT 		"gnuplot"
#endif /*!GNUPLOT*/

#ifndef PEAK_LIMIT
# define PEAK_LIMIT 		5.0
#endif /*!PEAK_LIMIT*/

/* "Average" values per plot. */
#ifndef AVG_DIVISOR
# define AVG_DIVISOR		100
#endif /*!AVG_INTERVAL*/

#define US2MS(val) ((long double)(val) / 1000)
#define US2S(val) ((long double)(val) / 1000000)

/*-[types]-----------------------------------------------------------*/

struct node {
	unsigned long long cntrl;
	unsigned long long slave;
};

#define NODEPOOL_BUFFGROW	4096
struct nodepool {
	struct node *pool;
	unsigned long pos;
	unsigned long size;
};

/*-[function declarations]-------------------------------------------*/

/* None of these functions are in c99 (we compile with std=c99) so we
 * need to manually include them. */

extern int
mkstemp(char *template);

extern FILE *
popen(const char *cmd, const char *type);

extern int
pclose(FILE *stream);

/*-[variable definitions]--------------------------------------------*/

static int quiet = 0;
static FILE *plot = NULL;
static char *png = NULL;
static char *sfile = NULL;
static char *afile = NULL;
static char *gnuplot = GNUPLOT;
static double peak = PEAK_LIMIT;
static unsigned int divisor = AVG_DIVISOR;

/*-[function definitions]--------------------------------------------*/

/** Comparison function passed along to qsort. */
/*anon*/ int
qsort_cmp(const void *p1, const void *p2)
{
	unsigned long long v1, v2, v3;
	v1 = ((const struct node *)p1)->cntrl;
	v2 = ((const struct node *)p2)->cntrl;
	v3 = v1 - v2;
	return((int)v3);
}

/** @return FALSE on failure, TRUE on success. */
static int
do_graph(const char *png)
{
	int rc;
	rc = png == NULL ? 
		fprintf(plot, "set terminal x11 persist\n") :
		fprintf(plot, "set terminal png\n"
			"set output '%s'\n", png);
	if (rc == -1) {
		warn("failed write to plot: %s", gnuplot);
		return(0);
	}
	rc = fprintf(plot, 
		"set title 'overhead (%.2f sec peak, "
			"%u curve interval)'\n"
		"set grid\n"
		"set size 1, 1\n"
		"set origin 0, 0\n"
		"set multiplot\n"
		"set size 0.5, 0.5\n"
		"set origin 0, 0.5\n"
		"set xlabel 'processor time (sec)'\n"
		"set ylabel 'overhead factor'\n"
		"plot '%s' using 2:3 with dots "
			"title 'record entry'\n"
		"set size 0.5, 0.5\n"
		"set origin 0.5, 0.5\n"
		"set xlabel 'processor time (sec)'\n"
		"set ylabel 'raw overhead (sec)'\n"
		"plot '%s' using 2:4 with dots "
			"title 'record entry'\n"
		"set size 0.5, 0.5\n"
		"set origin 0, 0\n"
		"set xlabel 'processor time (sec)'\n"
		"set ylabel 'average overhead (sec)'\n"
		"plot '%s' using 1:2 with linespoints "
			"title 'record entry'\n"
		"set size 0.5, 0.5\n"
		"set origin 0.5, 0\n"
		"set xlabel 'processor time (sec)'\n"
		"set ylabel 'median overhead (sec)'\n"
		"plot '%s' using 1:3 with linespoints "
			"title 'record entry'\n"
			, peak, divisor, sfile, sfile, 
			afile, afile);
	if (rc == -1)
		warn("failed write to plot: %s", gnuplot);
	return(rc == -1 ? 0 : 1);
}

static int
do_avgfile(const struct nodepool *pool)
{
	FILE *f;
	struct node *db;
	unsigned long i, sz, j;
	unsigned long long avg, median;
	long long v1, v2;
	assert(afile != NULL);
	assert(pool != NULL);

	db = pool->pool;
	assert(db != NULL);
	assert(pool->pos > 0);

	if ((f = fopen(afile, "w")) == NULL) {
		warn("failed open file: %s", afile);
		return(0);
	}

	if ((sz = pool->pos / divisor) == 0) {
		divisor = pool->pos;
		sz = 1;
	}
	for (i = 0; i < divisor; i++) {
		for (avg = 0, j = i * sz; j < (i * sz) + sz; j++) {
			assert(db[j].cntrl < LLONG_MAX);
			assert(db[j].slave < LLONG_MAX);
			v1 = (long long)db[j].cntrl;
			v2 = (long long)db[j].slave;
			avg += (unsigned long)
				(v1 - v2 < 0 ? 0 : v1 - v2);
		}
		avg /= sz;
		v1 = (long long)db[j - (sz / 2)].cntrl;
		v2 = (long long)db[j - (sz / 2)].slave;
		median = v1 - v2 < 0 ? 0 : v1 - v2;
		if (fprintf(f, "%21.6Lf %21.6Lf %21.6Lf\n",
					US2S(db[j].slave),
					US2S(avg), US2S(median))
				== -1) {
			warn("failed write to file: %s", afile);
			if (fclose(f) == -1)
				warn("failed close file: %s", afile);
			return(0);
		}
	}
	if (fclose(f) == -1) {
		warn("failed close file: %s", afile);
		return(0);
	}
	return(1);
}

/** @return FALSE on failure, TRUE on success. */
static int
do_scatterfile(const struct nodepool *pool)
{
	FILE *f;
	struct node *db;
	unsigned long i;
	long long v1, v2, v3;
	assert(sfile != NULL);
	assert(pool != NULL);

	db = pool->pool;
	assert(db != NULL);
	assert(pool->pos > 0);

	if ((f = fopen(sfile, "w")) == NULL) {
		warn("failed open file: %s", sfile);
		return(0);
	}
	for (i = 0; i < pool->pos; i++) {
		assert(db[i].cntrl < LLONG_MAX);
		assert(db[i].slave < LLONG_MAX);
		v1 = (long long)db[i].cntrl;
		v2 = (long long)db[i].slave;
		v3 = v1 - v2 < 0 ? 0 : v1 - v2;
		/* Check if we exceed the normalisation limit. */
		if ((long double)v3  / v2 > peak) {
			(void)fprintf(stderr, "Omission (exceeds "
					"threshold %.2f): index %lu, "
					"control %.4Lf ms, compute "
					"%.4Lf, factor %Lf.\n",
					peak, i, 
					US2S(v1), US2S(v2),
					(long double)v3 / v2);
			continue;
		}
		if (fprintf(f, "%21.6Lf %21.6Lf %21Lf %21Lf\n",
					US2S(v1), US2S(v2),
					(long double)v3 / v2,
					US2S(v3))
				== -1) {
			warn("failed write to file: %s", sfile);
			if (fclose(f) == -1)
				warn("failed close file: %s", sfile);
			return(0);
		}
	}
	if (fclose(f) == -1) {
		warn("failed close file: %s", sfile);
		return(0);
	}
	return(1);
}

static void
pool_resize(struct nodepool *pool, unsigned int addsize)
{
	unsigned int norm_addsz;
	assert(pool != NULL);
	assert(addsize > 0);
	if (pool->pos + addsize < pool->size)
		return;
	norm_addsz = addsize >= NODEPOOL_BUFFGROW ?
		addsize + NODEPOOL_BUFFGROW : NODEPOOL_BUFFGROW;
	pool->pool = realloc(pool->pool, 
			(pool->size + norm_addsz) *
			sizeof(struct node));
	assert(pool->pool != NULL);
	pool->size = pool->size + norm_addsz;
}

static void
doident(const struct master *master, const struct io_source *src,
		size_t size, const struct io_source *cntrl, 
		struct nodepool *pool)
{
	unsigned int args1, args2;
	unsigned long id;
	size_t i, j;
	struct node *db;
	assert(src != NULL);
	assert(master != NULL);
	assert(cntrl != NULL);

	id = master->id;

	/* Determine the number of control per-index entries. */
	for (args1 = 0, i = 0; i < cntrl->elems; i++)
		if (cntrl->data.cntrl[i].id == id)
			args1++;
	/* Determine the number of slave per-index entries. */
	for (args2 = 0, i = 0; i < size; i++)
		if (src[i].type == IO_SOURCE_SLAVE)
			for (j = 0; j < src[i].elems; j++) 
				if (src[i].data.slave[j].id == id) 
					args2++;
	/* We've not yet gotten all the entries flushed. */
	if (args1 != args2) 
		return;

	pool_resize(pool, args1);
	db = &pool->pool[pool->pos];
	pool->pos += args1;

	for (i = 0; i < cntrl->elems; i++)
		if (cntrl->data.cntrl[i].id == id)
			db[cntrl->data.cntrl[i].index].cntrl =
				cntrl->data.cntrl[i].wallc == 0 ?
				1 : cntrl->data.cntrl[i].wallc;
	for (i = 0; i < size; i++)
		if (src[i].type == IO_SOURCE_SLAVE)
			for (j = 0; j < src[i].elems; j++) {
				if (src[i].data.slave[j].id != id) 
					continue;
				/* Also normalise data. */
				db[src[i].data.slave[j].index].slave
					= src[i].data.slave[j].wallc
					== 0 ? 1 :
					src[i].data.slave[j].wallc;
			}
}

/*anon*/ int
doupdate(const struct io_source *src, size_t size)
{
	const struct io_source *master, *cntrl;
	struct nodepool pool;
	size_t i;
	assert(src != NULL);
	assert(size >= 3);

	/* First find the master and control nodes. */
	master = cntrl = NULL;
	for (i = 0; i < size; i++)
		if (src[i].type == IO_SOURCE_MASTER) 
			master = &src[i];
		else if (src[i].type == IO_SOURCE_CNTRL)
			cntrl = &src[i];
	assert(master != NULL);
	assert(cntrl != NULL);

	(void)memset(&pool, 0, sizeof(pool));

	/* Iterate through the completed jobs. */
	for (i = 0 ; i < master->elems; i++)
		doident(&master->data.master[i], 
				src, size, cntrl, &pool);

	if (pool.pos == 0)
		errx(1, "no measurements in data pool");

	/* Sort entries */
	qsort(pool.pool, pool.pos, sizeof(struct node), 
			qsort_cmp);

	if ( ! do_scatterfile(&pool)) {
		free(pool.pool);
		return(-1);
	}
	if ( ! do_avgfile(&pool)) {
		free(pool.pool);
		return(-1);
	}
	free(pool.pool);
	if ( ! quiet && ! do_graph(NULL))
		return(-1);
	if (png && ! do_graph(png))
		return(-1);
	return(0);
}

static void
usage(int code)
{
	extern char *__progname;
	(void)fprintf(stderr, "Usage: %s [-q] [-g gnuplot] "
			"[-f output] [-t threshold] "
			"[-c curve] file ...\n", 
			__progname);
	exit(code);
	/* NOTREACHED */
}

int
main(int argc, char *argv[])
{
	int c, rc, keep;
	const char *errstr;
	struct io_source src[MAX_SOURCES + 2];
	char tmp1[] = "/tmp/scatter.XXXXXXXXXX",
	     tmp2[] = "/tmp/avg.XXXXXXXXXX", *ep;
	
	(void)memset(src, 0, sizeof(src));
	keep = 0;

	while ((c = getopt(argc, argv, "c:g:f:qt:k")) != -1) 
		switch (c) {
		case('k'):
			keep = 1;
			break;
		case('g'):
			gnuplot = optarg;
			break;
		case('f'):
			png = optarg;
			break;
		case('q'):
			quiet = 1;
			break;
		case('c'):
			divisor = (unsigned int)strtonum
				(optarg, 1, UINT_MAX, &errstr);
			if (errstr != NULL)
				errx(1, "curve interval is %s: %s",
						errstr, optarg);
			break;
		case('t'):
			errno = 0;
			peak = strtod(optarg, &ep);
			if (optarg[0] == '\0' || *ep != '\0')
				errx(1, "invalid number: -t %s", 
						optarg);
			if (errno == ERANGE || peak == HUGE_VAL)
				errx(1, "invalid number: -t %s", 
						optarg);
			if (peak < 0.0)
				errx(1, "threshold must be positive"
						": -t %s", optarg);
			break;
		default:
			usage(1);
			/* NOTREACHED */
		}
	argc -= optind;
	argv += optind;

	if (mkstemp(tmp1) == -1)
		err(1, "failed mktemp: %s", tmp1);
	sfile = tmp1;
	if (mkstemp(tmp2) == -1)
		err(1, "failed mktemp: %s", tmp2);
	afile = tmp2;

	rc = 1;

	if ((plot = popen(gnuplot, "w")) == NULL) 
		goto out;

	if ( ! pl_mpi_prof_init(src, MAX_SOURCES + 2,
				doupdate, argc, argv))
		goto out;
	rc = pl_mpi_prof_run(src, argc, 0) ? 0 : 1;

out:	/* Egress point. */
	if (plot && fprintf(plot, "quit\n") == -1)
		warn("failed write to plot: %s", gnuplot);
	if (plot && pclose(plot) == -1)
		warn("failed pclose: %s", gnuplot);
	if ( ! keep) {
		if (sfile && unlink(sfile) == -1)
			warn("failed unlink: %s", sfile);
		if (afile && unlink(afile) == -1)
			warn("failed unlink: %s", afile);
	}
	return(rc);
}

