/* $Id: master.c,v 1.30 2007-01-24 12:11:52 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <err.h>
#include <limits.h>
#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <SWI-Prolog.h>

#include "pl_mpi.h"

/*-[types]-----------------------------------------------------------*/

struct job_context {
	unsigned long id;
	void *ptr;
	struct timeval start;
	struct timeval end;
};

/*-[function declarations]-------------------------------------------*/

/** Called by prolog to indicate the predicate has been called.  Has
 * nondeterministic bindings.
 * @param pred The predicate term.
 * @param in The input list.
 * @param out_result Result for index at "out_index".
 * @param out_yesno Whether the predicate failed ("no") or succeeded.
 * @param out_index Index in list of result "out_result".
 * @param handle Reentrant context. */
foreign_t
pl_dist(term_t pred, term_t in, term_t out_result, term_t out_yesno,
		term_t out_index, control_t handle);

/** A deterministic prolog predicate to load a file. 
 * @param file The filename term. */
foreign_t
pl_load(term_t file);

/** Called by all nodes (via prolog) upon exit.
 * @param code The exit code.
 * @param args Always set to NULL. */
void
pl_exit(int code, void *args);

/** Cut the predicate: cancel the job.
 * @param ctx The job context.
 * @param reent Whether we're dropping into this from above (failed
 * PL_REDO) or from prolog itself. 
 * @return The exit condition. */
static foreign_t
pl_dist_cut(struct job_context *ctx, int reent);

/** Process a "redo" request.  Blocks on solutions from control.
 * @param out_result Prolog "result" term.
 * @param out_index Prolog "index" term (index in argument list).
 * @param out_yesno Whether the result is valid or not ("no" or "yes").
 * @param ctx The job context.
 * @return -1 on failure, 0 on success. */
static int
pl_dist_redo(term_t out_result, term_t out_yesno, 
		term_t out_index, struct job_context *ctx);

/** Process a first call.  Triggers control to begin distributing. 
 * @param p_in Predicate name term.
 * @param in List of entities. 
 * @return A job context. */
static struct job_context *
pl_dist_start(term_t p_in, term_t in);

/*-[variable definitions]--------------------------------------------*/

/*-[function definitions]--------------------------------------------*/

static void *
_PL_get_pointer(term_t in)
{
	void *ctx;
	int64_t tmp;
	char *p;
	const char *errstr;
	switch(PL_term_type(in)) {
	case(PL_VARIABLE):
		/* FALLTHROUGH */
	case(PL_ATOM):
		/* FALLTHROUGH */
	case(PL_INTEGER):
		/* FALLTHROUGH */
	case(PL_FLOAT):
		if ( ! PL_get_chars(in, &p, CVT_ALL)) {
			MPI_WARN("failed to convert input characters");
			return(NULL);
		}
		tmp = (int64_t)strtonum
			(p, INT64_MIN, INT64_MAX, &errstr);
		if (errstr) {
			MPI_WARN("input value is %s: %s", errstr, p);
			return(NULL);
		}
		break;
	default:
		MPI_WARN("unknown input type %d (known: atom, "
				"variable, integer, float)", 
				PL_term_type(in));
		return(NULL);
	}
	(void)memcpy(&ctx, &tmp, sizeof(void *));
	return(ctx);
}

foreign_t
pl_load(term_t file)
{
	int rc, status;
	MPI_Request rq;
	unsigned int len, i;
	char *fname, c;
	fid_t fid;
	term_t term;
	predicate_t pred;

	if ( ! PL_get_atom_chars(file, &fname)) {
		MPI_WARN("failed to get atom characters");
		PL_fail;
	}
	len = strlen(fname);

	MPI_DEBUG("consult request: %s", fname);

	/* Issue the event to all compute nodes. */
	c = TAG_EVENT_LOAD;
	for (i = 2; i < nodes; i++) {
		MPI_ISEND(&c, 1, MPI_CHAR, i, TAG_EVENT, &rq);
		MPI_ISEND(&len, 1, MPI_UNSIGNED, i, TAG_LOAD_FLEN, &rq);
		MPI_ISEND(fname, len, MPI_CHAR, i, TAG_LOAD_FNAME, &rq);
	}

	/* Gather status from all compute nodes. */
	for (status = 0, i = 2; i < nodes; i++) {
		MPI_RECV(&rc, 1, MPI_INT, i, TAG_LOAD_RESULT);
		if (rc != TRUE)
			MPI_DEBUG("node fails consult: %d, %s", 
					i, fname);
		status += (rc == TRUE ? 0 : 1);
	}
	/* If any one compute node has failed, fail also. */
	if (status > 0)
		PL_fail;

	fid = PL_open_foreign_frame();

	term = PL_new_term_refs(1);
	PL_put_atom_chars(term, fname);
	pred = PL_predicate("consult", 1, "user");
	status = PL_call_predicate(NULL, PL_Q_NORMAL, pred, term);

	PL_discard_foreign_frame(fid);

	/* Return whether we succeeded. */
	if (status == TRUE)
		PL_succeed;
	PL_fail;
}

void
pl_exit(int code, void *args)
{
	char c;
	unsigned int i;
	MPI_Request rq;
	(void)args;

	c = TAG_EVENT_HALT;
	for (i = 1; i < nodes; i++) {
		MPI_ISEND(&c, 1, MPI_CHAR, i, TAG_EVENT, &rq);
		MPI_ISEND(&code, 1, MPI_INT, i, TAG_HALT_CODE, &rq);
	}

	MPI_BARRIER();
	MPI_FINALIZE();
}

foreign_t
pl_dist_cut(struct job_context *ctx, int reent)
{
	char code;
	MPI_Request rq;
	struct timeval tv;
	unsigned long long tt;
	assert(ctx != NULL);
	assert(ctx->ptr != NULL);

	code = TAG_EVENT_CUT;
	MPI_ISEND(&code, 1, MPI_CHAR, RANK_CONTROL, TAG_EVENT, &rq);
	MPI_ISEND(&ctx->ptr, (int)sizeof(void *), MPI_CHAR, 
			RANK_CONTROL, TAG_CUT_HANDLE, &rq);

	if (gettimeofday(&ctx->end, NULL) == -1)
		MPI_ABORT();

	timersub(&ctx->end, &ctx->start, &tv);
	/* This can be less than zero on some systems. */
	if (tv.tv_sec < 0)
		tv.tv_sec = tv.tv_usec = 0;
	tv.tv_usec = tv.tv_usec < 0 ? 0 : tv.tv_usec;
	/* Cast aggr. to larger-width type. */
	tt = ((unsigned long long)tv.tv_sec * 1000000) +
		(unsigned long long)tv.tv_usec;
	if (opts.dump != NULL) {
		if (fprintf(opts.dump, 
				"%10p %10lu %20llu\n", 
				ctx, ctx->id, tt) == -1)
			MPI_ABORT();
		if (fflush(opts.dump) == -1)
			MPI_ABORT();
	}

	free(ctx);

	/* Fail if we're coming from a failed predicate, above. */
	if (reent)
		PL_fail;
	PL_succeed;
}

int
pl_dist_redo(term_t out_result, term_t out_yesno, term_t out_index, 
		struct job_context *ctx)
{
	int ind, yesno, type;
	unsigned int size;
	MPI_Request rq;
	char code, *result, *ep;
	double farg;
	long larg;
	assert(ctx != NULL);
	assert(ctx->ptr != NULL);

	/* Trigger the data query. */
	code = TAG_EVENT_REDO;
	MPI_ISEND(&code, 1, MPI_CHAR, RANK_CONTROL, TAG_EVENT, &rq);
	
	/* Send the relevant pending-data handle. */
	MPI_ISEND(&ctx->ptr, (int)sizeof(void *), MPI_CHAR, 
			RANK_CONTROL, TAG_REDO_HANDLE, &rq);

	/* Retrieve type, out_index and out_result. */
	MPI_RECV(&ind, 1, MPI_INT, RANK_CONTROL, TAG_REDO_INDEX);
	/* Return immediately if we're "false". */
	if (ind == RESULT_FAILURE)
		return(-1);

	/* Unify the output types. */
	if ( ! PL_unify_integer(out_index, ind))
		MPI_ABORT();

	MPI_RECV(&yesno, 1, MPI_INT, RANK_CONTROL, TAG_REDO_RESULTYES);
	if ( ! PL_unify_integer(out_yesno, yesno))
		MPI_ABORT();

	/* If the predicate returned FALSE then return immediately; no
	 * valid result is following. */
	if (yesno) {
		MPI_RECV(&type, 1, MPI_INT, RANK_CONTROL, 
				TAG_REDO_RESULTTYPE);
		MPI_RECV(&size, 1, MPI_UNSIGNED, RANK_CONTROL, 
				TAG_REDO_RESULTSIZE);
		result = (char *)calloc(size + 1, 1);
		assert(result != NULL);
		MPI_RECV(result, size, MPI_CHAR, RANK_CONTROL,
				TAG_REDO_RESULT);

		MPI_DEBUG("affirmative %p (%lu) (type %d size %d)", 
				ctx->ptr, ctx->id, type, size);

		switch(type) {
		case(PL_ATOM):
			if ( ! PL_unify_atom_chars(out_result, result))
				MPI_ABORT();
			break;
		case(PL_STRING):
			if ( ! PL_unify_string_nchars(out_result, 
						size, result))
				MPI_ABORT();
			break;
		case(PL_INTEGER):
			larg = strtol(result, &ep, 10);
			if (result[0] == '\0' || *ep != '\0')
				MPI_ABORT();
			if (larg == LONG_MAX || larg == LONG_MIN)
				MPI_ABORT();
			if ( ! PL_unify_integer(out_result, larg))
				MPI_ABORT();
			break;
		case(PL_FLOAT):
			farg = strtod(result, &ep);
			if (result[0] == '\0' || *ep != '\0')
				MPI_ABORT();
			if (farg == HUGE_VAL || farg == -HUGE_VAL)
				MPI_ABORT();
			if ( ! PL_unify_float(out_result, farg))
				MPI_ABORT();
			break;
		default:
			MPI_WARN("unknown output type %d (known: "
					"atom, string, integer, "
					"float)", type);
			MPI_ABORT();
		}
		free(result);
	} else
		MPI_DEBUG("failure %p (%lu)", ctx->ptr, ctx->id);
	return(0);
}

struct job_context *
pl_dist_start(term_t p_in, term_t in)
{
	char *pred, *arg, code;
	int type;
	term_t head, list;
	unsigned int size, len;
	MPI_Request rq;
	struct job_context *ctx;

	/* Allocate context for this job. */
	ctx = calloc(1, sizeof(struct job_context));
	assert(ctx != NULL);

	{ /* Get the function. */
		functor_t p;
		if ( ! PL_get_functor(p_in, &p)) {
			warnx("failed to get functor");
			return(NULL);
		}
		/* XXX Discard const type! */
		/* LINTED */
		pred = (char *)PL_atom_chars(PL_functor_name(p));
		assert(pred != NULL);
	}

	/* Notify that we're sending data. */
	code = TAG_EVENT_FIRSTCALL;
	MPI_ISEND(&code, 1, MPI_CHAR, RANK_CONTROL, TAG_EVENT, &rq);

	/* Send the function name. */
	size = strlen(pred);
	MPI_ISEND(&size, 1, MPI_UNSIGNED, RANK_CONTROL, 
			TAG_FIRSTCALL_PREDSIZE, &rq);	
	MPI_ISEND(pred, (int)size, MPI_CHAR, RANK_CONTROL,
			TAG_FIRSTCALL_PRED, &rq);

	/* Send the argument array. */
	head = PL_new_term_ref();

	/* Length of argument array. */
	for (len = 0, list = PL_copy_term_ref(in); 
			PL_get_list(list, head, list); len++) 
		; /* Do nothing */
	MPI_ISEND(&len, 1, MPI_UNSIGNED, RANK_CONTROL,
			TAG_FIRSTCALL_COUNT, &rq);

	/* Argument array itself.  We have to determine the type for
	 * each input element manually and copy it into the input
	 * element type.  Note that at this time we don't support
	 * anything that's not an atomic type, string or number (like a
	 * function, or a variable). */
	
	MPI_DEBUG("sending job (%s, %u slots)", pred, len);

	for (list = PL_copy_term_ref(in);
			PL_get_list(list, head, list); ) {
		type = PL_term_type(head);
		switch(type) {
		case(PL_ATOM):
			/* FALLTHROUGH */
		case(PL_INTEGER):
			/* FALLTHROUGH */
		case(PL_FLOAT):
			if (PL_get_chars(head, &arg, CVT_ALL | 
						REP_UTF8) == FALSE)
				MPI_ABORT();
			size = strlen(arg);
			break;
		case(PL_STRING):
			if (PL_get_string_chars(head, &arg, &size) 
					== FALSE)
				MPI_ABORT();
			break;
		case(PL_TERM):
			/* FALLTHROUGH */
		case(PL_VARIABLE):
			/* FIXME: this stalls the system. */
			MPI_WARN("unknown input type %d (known: "
					"atom, string, integer, "
					"float)", type);
			return(NULL);
		}
		MPI_ISEND(&type, 1, MPI_UNSIGNED, RANK_CONTROL,
				TAG_FIRSTCALL_DATATYPE, &rq);
		MPI_ISEND(&size, 1, MPI_UNSIGNED, RANK_CONTROL,
				TAG_FIRSTCALL_DATASIZE, &rq);
		MPI_ISEND(arg, size, MPI_CHAR, RANK_CONTROL,
				TAG_FIRSTCALL_DATA, &rq);
	}

	/* Receive the job context. */
	MPI_RECV(&ctx->ptr, (int)sizeof(void *), MPI_CHAR, 
			RANK_CONTROL, TAG_FIRSTCALL_HANDLE);
	MPI_RECV(&ctx->id, 1, MPI_UNSIGNED_LONG, RANK_CONTROL, 
			TAG_FIRSTCALL_IDENT);

	MPI_DEBUG("registered job (%p, %lu)", 
			ctx->ptr, ctx->id);
	if (gettimeofday(&ctx->start, NULL) == -1)
		MPI_ABORT();
	return(ctx);
}

/*anon*/ foreign_t
pl_aquery(term_t in, term_t out)
{
	struct job_context *ctx;
	long count;
	char event;
	MPI_Request rq;

	if ((ctx = _PL_get_pointer(in)) == NULL)
		PL_fail;
	event = TAG_EVENT_QUERY;
	MPI_ISEND(&event, 1, MPI_CHAR, RANK_CONTROL, 
			TAG_EVENT, &rq);
	MPI_ISEND(&ctx->ptr, (int)sizeof(void *), MPI_CHAR,
			RANK_CONTROL, TAG_QUERY_HANDLE, &rq);
	MPI_RECV(&count, 1, MPI_LONG, RANK_CONTROL,
			TAG_QUERY_RESULT);
	if ( ! PL_unify_integer(out, count))
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t
pl_adist(term_t pred, term_t in, term_t out)
{
	struct job_context *ctx;
	/* NB.  We need to put this into a form easily cast back into a
	 * pointer.  Since we want to be compatible with both 64- and
	 * 32-bit machines, make sure our pointer fits into an
	 * acceptably-sized buffer. */
	if ((ctx = pl_dist_start(pred, in)) == NULL)
		PL_fail;
	MPI_DEBUG("asynchronous query (%p)", ctx);
	if ( ! PL_unify_int64(out, (int64_t)ctx)) {
		MPI_ABORT();
	}
	PL_succeed;
}

/*anon*/ foreign_t
pl_aget(term_t in, term_t out_result, term_t out_yesno,
		term_t out_index)
{
	struct job_context *ctx;
	if ((ctx = _PL_get_pointer(in)) == NULL)
		PL_fail;
	/* Indicates no results available. */
	if (pl_dist_redo(out_result, out_yesno, out_index, ctx)
			== -1)
		PL_fail;
	PL_succeed;
}

/*anon*/ foreign_t
pl_aclose(term_t in)
{
	struct job_context *ctx;
	if ((ctx = _PL_get_pointer(in)) == NULL)
		PL_fail;
	(void)pl_dist_cut(ctx, 0/*not important*/);
	PL_succeed;
}

/*anon*/ foreign_t
pl_dist(term_t pred, term_t in, term_t out_result, term_t out_yesno,
		term_t out_index, control_t handle)
{
	struct job_context *ctx;
	int code;

	ctx = NULL;
	code = 0;
	switch(PL_foreign_control(handle)) {
	case(PL_FIRST_CALL):
		/* We want this to fall through to the redo statement,
		 * because FIRST_CALL is supposed to return a result. */
		if ((ctx = pl_dist_start(pred, in)) == NULL)
			PL_fail;
		/* FALLTHROUGH */
	case(PL_REDO):
		/* If we're coming from above, our context is already
		 * initialised, else get it. */
		if (ctx == NULL)
			ctx = PL_foreign_context_address(handle);
		/* If we've no more solutions, drop down and cancel the
		 * running predicate. */
		if (pl_dist_redo(out_result, out_yesno, out_index, 
					ctx) == 0)
			PL_retry_address(ctx);
		code = 1;
		/* FALLTHROUGH */
	case(PL_CUTTED):
		/* If we're coming from above, our context is already
		 * initialised, else get it. */
		if (ctx == NULL)
			ctx = PL_foreign_context_address(handle);
		return(pl_dist_cut(ctx, code));
	}

	/* We shouldn't get here. */
	MPI_ABORT();
	/* NOTREACHED */
}

/*extern*/ void
init_prolog(int argc, char *argv[])
{
	static PL_extension pred[] = {
		{ PREDICATE_DIST, 5, pl_dist, PL_FA_NONDETERMINISTIC },
		{ PREDICATE_LOAD, 1, pl_load, 0 },
		{ PREDICATE_ADIST, 3, pl_adist, 0 },
		{ PREDICATE_AGET, 4, pl_aget, 0 },
		{ PREDICATE_ACLOSE, 1, pl_aclose, 0 },
		{ PREDICATE_AQUERY, 2, pl_aquery, 0 },
		{ NULL, 0, NULL, 0 }
	};
	assert(argv != NULL);
	assert(argc >= 1);

	if ( ! PL_initialise(argc, argv))
		MPI_ABORT();

	/* Only the root ("master") needs to register hooks. */

	if (rank != 0)
		return;

	/* Prolog initialisation. */
	PL_register_extensions_in_module("user", pred);
	/* Register our exit handler. */
	PL_on_halt(pl_exit, NULL);
}

/*extern*/ void
init_mpi(int *argc, char ***argv)
{
	int tmp;
	MPI_INIT(argc, argv);
	MPI_COMM_RANK(&tmp);
	assert(tmp >= 0);
	rank = tmp;
	MPI_COMM_SIZE(&tmp);
	assert(tmp > 0);
	nodes = tmp;

	if (nodes > 2)
		return;

	if (rank == 0)
		warnx("at least three nodes required (have %u)",
			nodes);
	MPI_FINALIZE();
	exit(1);
	/* NOTREACHED */
}

