/* $Id: pl_mpi.h,v 1.29 2007-01-22 21:38:18 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef PL_MPI_H
#define PL_MPI_H

__BEGIN_DECLS

/* Needed by all header files. */
#include "compat.h"

/* Names for predicates. */
#define PREDICATE_DIST		"mpidist"
#define PREDICATE_ADIST		"ampidist"
#define PREDICATE_AGET		"ampiget"
#define PREDICATE_ACLOSE	"ampiclose"
#define PREDICATE_AQUERY	"ampiquery"
#define PREDICATE_LOAD		"mpiload"

/** Passed from control to master (in lieu of an index) to indicate that
 * no more values are available (backtracking must cease). */
#define RESULT_FAILURE		(-1)

/** Structure defining a predicate's job.  This encompasses all
 * arguments passed along, status, and so forth.  Notice that jobs can
 * have varying types of input and output, although these are restricted
 * to native types. */
struct cntrl_pending
{
	unsigned long ident;
	/**< The job identifier (unique to each structure). */
	unsigned int count;
	/**< Number of entries (element slots) in ``data''. */
	char *pred;
	/**< Predicate name. */
	unsigned int predlen;
	/**< Length of predicate name string. */

	/** Data pertaining to each input/output slot (that is, data
	 * distributed and gathered per predicate element). */
	struct params {
		int input_type;
		/**< Type of input (PL_ATOM, etc.). */
		char *input;
		/**< String representation of input. */
		unsigned int input_length;
		/**< Length of string in ``input''. */
	
		int output_type;
		/**< Type of output (PL_ATOM, etc.). */
		char *output;
		/**< String representation of output (may be NULL if
		 * this output entry was never gathered). */
		unsigned int output_length;
		/**< Length of string in ``output'' (may be a random
		 * value if output is unset. */
		int output_yesno;
		/**< TRUE if the predicate succeeded, FALSE if it
		 * failed. */
	
		struct timeval start;
		/**< Time-stamp at pre-distribution. */
		struct timeval end;
		/**< Time-stamp at post-collection. */

		unsigned int node;
		/**< Node (rank) upon which the job ran. */

/** Has been created, but not yet distributed. */
#define STATUS_NEW		0
/** Has been distributed and is pending on a node. */
#define STATUS_WAITING		1
/** Has been returned to control but has not been collected by the
 * master node. */
#define STATUS_COMPLETE		2
/** Has been collected by the master node. */
#define STATUS_CONFIRMED	3
		short status;
		/**< Status indicator of this entry. */
	} *data;

/** The master is blocking, waiting for an entry in this predicate to be
 * returned to it. */
#define FLAGS_BLOCK		0x01
/** The entire entry is no longer valid. */
#define FLAGS_CUTTED		0x02
	uint8_t flags;
	/**< Miscellaneous flags set on the structure. */
};

/** Options are global to pl-mpi and indicate various elements of the
 * run-time system. */
struct options
{
	FILE *trace;
	/**< Stream for trace/debug file.  If NULL, no tracing should
	 * occur. */
	char trace_file[MAXPATHLEN];
	/**< The underlying file for ``trace''. */
	FILE *dump;
	/**< Stream for dump/profile file.  If NULL, no profiling should
	 * occur. */
	char dump_file[MAXPATHLEN];
	/**< The underlying file for ``dump''. */
} opts;

/** Main listening loop for the control node.
 * @return 0 on failure, 1 on success. */
extern int
loop_cntrl(void);

/** Main listening loop for slave nodes.Z
 * why the parameter isn't constified).
 * @return 0 on failure, 1 on success. */
extern int
loop_slave(void);

/** Initialise the MPI subsystem.  Distributes arguments to all running
 * systems.  This should be called before any argument processing. */
extern void
init_mpi(int *argc, char ***argv);

/** Initialise the Prolog subsystem. */
extern void
init_prolog(int argc, char *argv[]);

/** Rank for the root ("master") node accepting prolog events. */
#define RANK_ROOT		0
/** Rank for the control node accepting events from the master node and
 * slave processors. */
#define RANK_CONTROL		1

/* Tags for MPI communication between nodes. */
#define TAG_EVENT		0
#define TAG_COMPLETE_RESULT	1
#define TAG_FIRSTCALL_COUNT	2
#define TAG_FIRSTCALL_DATASIZE	3
#define TAG_FIRSTCALL_DATA	4
#define TAG_FIRSTCALL_HANDLE	5
#define TAG_FIRSTCALL_PRED	6
#define TAG_FIRSTCALL_PREDSIZE	7
#define TAG_REDO_HANDLE		8
#define TAG_REDO_RESULT		9
#define TAG_REDO_INDEX		10
#define TAG_WORKER_PREDSIZE	11
#define TAG_WORKER_PRED		12
#define TAG_WORKER_DATASIZE	13
#define TAG_WORKER_DATA		14
#define TAG_CUT_HANDLE		15
#define TAG_LOAD_FLEN		16
#define TAG_LOAD_FNAME		17
#define TAG_FIRSTCALL_DATATYPE	18
#define TAG_WORKER_DATATYPE	19
#define TAG_REDO_RESULTTYPE	20
#define TAG_WORKER_RESULTTYPE	21
#define TAG_FIRSTCALL_RESULTTYPE 22
#define TAG_REDO_RESULTYES	23
#define TAG_COMPLETE_YESNO	24
#define TAG_COMPLETE_RESULTSIZE	25
#define TAG_COMPLETE_RESULTTYPE	26
#define TAG_REDO_RESULTSIZE	27
#define TAG_HALT_CODE		28
#define TAG_LOAD_RESULT		29
#define TAG_WORKER_PREDKEY	30
#define TAG_WORKER_PREDINDEX	31
#define TAG_FIRSTCALL_IDENT	32
#define TAG_QUERY_RESULT	33
#define TAG_QUERY_HANDLE	34

/* Event tags (signal an event type). */
#define TAG_EVENT_FIRSTCALL	1
#define TAG_EVENT_REDO		2
#define TAG_EVENT_WORKER	3
#define TAG_EVENT_COMPLETE	4
#define TAG_EVENT_HALT		5
#define TAG_EVENT_CUT		6
#define TAG_EVENT_LOAD		7
#define TAG_EVENT_QUERY		8

/** Our rank in the MPI group. */
extern unsigned int
rank;

/** Number of participating nodes. */
extern unsigned int
nodes;

/*---------------------------------------------------------------------
 * MPI components follow.
 *-------------------------------------------------------------------*/

extern void
mpidbg(const char *file, int line, const char *fmt, ...)
	__attribute__ ((format (printf, 3, 4)));

extern void
mpitrace(const char *fmt, ...)
	__attribute__ ((format (printf, 1, 2)));

#define MPI_WARN(fmt, ...) do { \
	mpidbg(__FILE__, __LINE__, fmt, ##__VA_ARGS__); \
	warnx(fmt, ##__VA_ARGS__); } while(0)

#define MPI_DEBUG(fmt, ...) \
	mpidbg(__FILE__, __LINE__, fmt, ##__VA_ARGS__)

#define MPI_TRACE(fmt, ...) \
	mpitrace(fmt, ##__VA_ARGS__)

#define MPI_INIT(x1, x2) do \
	if (MPI_Init((x1), (x2)) != MPI_SUCCESS) \
		MPI_ABORT(); while(0)

#define MPI_COMM_RANK(x1) do \
	if (MPI_Comm_rank(MPI_COMM_WORLD, (x1)) != MPI_SUCCESS) \
		MPI_ABORT(); while(0)

#define MPI_COMM_SIZE(x1) do \
	if (MPI_Comm_size(MPI_COMM_WORLD, (x1)) != MPI_SUCCESS) \
		MPI_ABORT(); while(0)

#define MPI_BARRIER() do \
	if (MPI_Barrier(MPI_COMM_WORLD) != MPI_SUCCESS) \
		MPI_ABORT(); while(0)

#define MPI_FINALIZE() (void)MPI_Finalize()

/** Trigger an abort.  Please note that this does not call through to
 * MPI_Abort.  Instead, it triggers an abort with abort(3) in order to
 * properly generate a coredump for post-mortem analysis. */
#define MPI_ABORT() do { \
	(void)fprintf(stderr, "User-level abort from " \
			"%d: %s:%d (%s)\n", \
			rank, __FILE__, __LINE__, __func__); \
	(void)fflush(stderr); \
	abort(); } while(0)

#define MPI_RECV_ST(x1, x2, x3, x4, x5, x7) do { \
	if (MPI_Recv((x1),(x2),(x3),(x4),(x5), \
				MPI_COMM_WORLD,(x7)) != MPI_SUCCESS) \
		MPI_ABORT(); } while(0)

#define MPI_RECV(x1, x2, x3, x4, x5) do { \
	if (MPI_Recv((x1),(x2),(x3),(x4),(x5), MPI_COMM_WORLD, \
				MPI_STATUS_IGNORE) != MPI_SUCCESS) \
		MPI_ABORT(); } while(0)

#define MPI_SEND(x1, x2, x3, x4, x5, x6) do { \
	if (MPI_Send((x1),(x2),(x3),(x4),(x5),(x6)) != MPI_SUCCESS) \
		MPI_ABORT(); } while(0)

#define MPI_ISEND(x1, x2, x3, x4, x5, x7) do { \
	if (MPI_Isend((x1),(x2),(x3),(x4),(x5), \
				MPI_COMM_WORLD,(x7)) != MPI_SUCCESS) \
		MPI_ABORT(); } while(0)

#define MPI_IPROBE(x1, x2, x3, x4, x5) do { \
	if (MPI_Iprobe((x1),(x2),(x3),(x4),(x5)) != MPI_SUCCESS) \
		MPI_ABORT(); } while(0)

#define MPI_PROBE(x1, x2, x4) do { \
	if (MPI_Probe((x1),(x2),MPI_COMM_WORLD,(x4)) != MPI_SUCCESS) \
		MPI_ABORT(); } while(0)

__END_DECLS

#endif /* ! PL_MPI_H */
