/* $Id: profile.c,v 1.8 2007-01-22 21:38:18 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "profile.h"

/*-[variable definitions]--------------------------------------------*/

static int (*update_p)(const struct io_source *src, size_t size);

/*-[function definitions]--------------------------------------------*/

static void
data_realloc(void **data, size_t scale_factor, size_t *scale_size,
		size_t elems, size_t ssize)
{
	assert(data != NULL);
	assert(scale_factor > 0);
	assert(scale_size != NULL);
	assert(elems <= *scale_size);
	if (elems != *scale_size)
		return;
	*data = realloc(*data, ssize * (*scale_size + scale_factor));
	assert(*data != NULL);
	*scale_size += scale_factor;
}

static int
doline(struct io_source *src, const char line[MAX_LINE])
{
	int want, have;
	size_t sz, fact, elems;
	void **ptr;
	assert(src != NULL);

	want = have = 0;
	if (line[0] == '#') {
		char pound[2];
		char header[5];
		(void)memset(pound, 0, sizeof(pound));
		(void)memset(header, 0, sizeof(header));
		if (strstr(line, "# rank") != NULL) {
			want = 3;
			have = sscanf(line, "%1s %4s %10u", 
					pound, header, &src->node);
			if (have != want) {
				warnx("failed parsing, %d fields, "
						"want %d: %s:%u",
						have, want, 
						src->file, src->line);
				return(0);
			}
			if (strcmp(pound, "#") != 0) {
				warnx("failed parsing, expected '#', "
						"have '%s': %s:%u",
						pound, src->file, 
						src->line);
				return(0);
			}
			if (strcmp(header, "rank") != 0) {
				warnx("failed parsing, expected 'rank'"
						" have '%s': %s:%u",
						header, src->file, 
						src->line);
				return(0);
			}
		}
		return(1);
	}

	/* First check reallocation. */
	switch(src->type) {
	case(IO_SOURCE_MASTER):
		ptr = (void **)&src->data.master;
		fact = MASTER_SCALE_SZ;
		sz = sizeof(struct master);
		break;
	case(IO_SOURCE_CNTRL):
		ptr = (void **)&src->data.cntrl;
		fact = CNTRL_SCALE_SZ;
		sz = sizeof(struct cntrl);
		break;
	case(IO_SOURCE_SLAVE):
		ptr = (void **)&src->data.slave;
		fact = SLAVE_SCALE_SZ;
		sz = sizeof(struct slave);
		break;
	default:
		abort();
	}

	elems = src->elems;
	data_realloc(ptr, fact, &src->scale_size, elems, sz);

	switch(src->type) {
	case(IO_SOURCE_MASTER):
		have = sscanf(line, MASTER_INPUT_FMT,
				&src->data.master[elems].ctx,
				&src->data.master[elems].id,
				&src->data.master[elems].wallc);
		want = 3;
		break;
	case(IO_SOURCE_CNTRL):
		have = sscanf(line, CNTRL_INPUT_FMT,
				&src->data.cntrl[elems].id,
				&src->data.cntrl[elems].index,
				&src->data.cntrl[elems].slave,
				&src->data.cntrl[elems].wallc);
		want = 4;
		break;
	case(IO_SOURCE_SLAVE):
		have = sscanf(line, SLAVE_INPUT_FMT,
				&src->data.slave[elems].id,
				&src->data.slave[elems].index,
				&src->data.slave[elems].cpu,
				&src->data.slave[elems].wallc);
		want = 4;
		break;
	}
	if (have != want) {
		warnx("failed parsing, %d fields, want %d: %s:%u",
				have, want, src->file, src->line);
		return(0);
	}
	src->elems++;
	return(1);
}

/** Drain all available data from a file.  Note that we must be
 * line-buffered and always end on a newline.  
 * @return -1 on failure or the number of processed entries. */
static int
dodrain(struct io_source *src)
{
	static char line[MAX_LINE];
	ssize_t sz, t, total;
	int entries;
	assert(src != NULL);

	total = 0;
	entries = 0;
line:
	for (t = 0; t < MAX_LINE; t++, total++) { 
		if ((sz = read(src->fd, &line[t], 1)) == -1) {
			warn("failed reading from file: %s:%u", 
					src->file, src->line);
			return(-1);
		} else if (sz == 0) 
			break;

		if (line[t] != '\n')
			continue;
		line[t] = '\0';
		if ( ! doline(src, line))
			return(-1);
		entries++;
		src->line++;
		goto line;
	}
	if (t >= MAX_LINE) {
		warnx("file line too long: %s:%u (%d max)", 
				src->file, src->line, MAX_LINE);
		return(-1);
	}
	if (total == 0 || t == 0)
		return(entries);
	warnx("file line not newline-buffered: %s:%u", 
			src->file, src->line);
	return(-1);
}

/*extern*/ int
pl_mpi_prof_init(struct io_source *src, size_t size,
		int (*update)(const struct io_source *src, 
			size_t size),
		int argc, char *argv[])
{
	int c, have_master, have_cntrl, have_slave;
	assert(src != NULL);
	assert(update != NULL);

	update_p = update;

	if (size < (size_t)argc) {
		warnx("array too small (%d, %zd)", argc, size);
		return(0);
	}

	/* Fill in our array of nodes. */
	have_master = have_cntrl = have_slave = 0;
	for (c = 0; c < argc; c++) {
		if (strstr(argv[c], "-0.txt") != NULL) {
			have_master++;
			src[c].type = IO_SOURCE_MASTER;
		} else if (strstr(argv[c], "-1.txt") != NULL) {
			have_cntrl++;
			src[c].type = IO_SOURCE_CNTRL;
		} else {
			have_slave++;
			src[c].type = IO_SOURCE_SLAVE;
		}
		if ((src[c].fd = open(argv[c], O_RDONLY)) == -1) {
			warn("failed open: %s", argv[c]);
			return(0);
		}
		src[c].file = argv[c];
		src[c].line = 1;
	}

	/* Make sure we've the correct number of nodes. */
	if (have_master != 1) {
		warnx("exactly one master must be provided (have %d)",
				have_master);
		return(0);
	}
	if (have_cntrl != 1) {
		warnx("exactly one control must be provided (have %d)",
				have_cntrl);
		return(0);
	}
	if (have_slave == 0) {
		warnx("at least one slave must be provided (have %d)",
				have_slave);
		return(0);
	}
	return(1);
}

/*extern*/ int
pl_mpi_prof_run(struct io_source *src, size_t size,
		unsigned int interval)
{
	unsigned int c;
	int count, tt;
	assert(src != NULL);

	/* First drain the existing data. */
	for (count = 0, c = 0; c < size; c++)
		if ((tt = dodrain(&src[c])) == -1)
			return(0);
		else 
			count += tt;

	if (count > 0 && (*update_p)(src, size) == -1)
		return(0);

	if (interval == 0)
		return(1);

	/* Main functioning loop. */
	for (;; /*Infinite loop.*/ ) {
		(void)sleep(interval);
		for (count = 0, c = 0; c < size; c++) 
			if ((tt = dodrain(&src[c])) == -1)
				return(0);
			else
				count += tt;
		if (count > 0 && (*update_p)(src, size) == -1)
			return(0);
	}

	/* NOTREACHED */
}

