/* $Id: slave.c,v 1.27 2007-01-22 21:38:18 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <assert.h>
#include <err.h>
#include <limits.h>
#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <SWI-Prolog.h>

#include "pl_mpi.h"

/*-[ function declarations ]-----------------------------------------*/

/** Calls through to _do_slave_work after recording state.
 * @param source The request source (node). */
static void
do_slave_work(int source);

/** Receive predicate information, call out ot the associated predicate
 * and then return our results appropriately. 
 * @param source The request source (node). */
static void
_do_slave_work(int source, struct timeval *t1, struct timeval *t2,
		struct rusage *ru1, struct rusage *ru2);

/** Calls through to _do_slave_load after recording state.
 * @param source The request source (node). */
static void
do_slave_load(int source);

/** Receive a file to load (a.k.a. use consult/1).
 * @param source The request source (node). */
static void
_do_slave_load(int source);

/*-[ function definitions ]------------------------------------------*/

void
do_slave_load(int source)
{
	fid_t fid;
	fid = PL_open_foreign_frame();
	_do_slave_load(source);
	PL_discard_foreign_frame(fid);
}

void
_do_slave_load(int source)
{
	unsigned int len;
	predicate_t pred;
	term_t term;
	char *fname;
	int rc;
	MPI_Request rq;

	MPI_RECV(&len, 1, MPI_UNSIGNED, source, TAG_LOAD_FLEN);
	fname = calloc(len + 1, 1);
	assert(fname != NULL);
	MPI_RECV(fname, len, MPI_CHAR, source, TAG_LOAD_FNAME);

	term = PL_new_term_refs(1);
	PL_put_atom_chars(term, fname);
	pred = PL_predicate("consult", 1, "user");

	rc = PL_call_predicate(NULL, PL_Q_NORMAL, pred, term);

	free(fname);

	MPI_ISEND(&rc, 1, MPI_INT, source, TAG_LOAD_RESULT, &rq);
}

void
do_slave_work(int source)
{
	struct timeval tv, tv1, tv2, wc1, wc2, wc;
	struct rusage ru1, ru2;
	unsigned long key;
	unsigned long long tt1, tt2;
	fid_t fid;
	unsigned int idex;

	/* NB: struct rusage is not defined in SUS, so be sure to check
	 * both BSD and Linux manuals before settling on output
	 * parameters to check. */

	MPI_RECV(&key, 1, MPI_UNSIGNED_LONG, 
			source, TAG_WORKER_PREDKEY);
	MPI_RECV(&idex, 1, MPI_UNSIGNED, 
			source, TAG_WORKER_PREDINDEX);

	MPI_TRACE("%10s %10lu %10u", "begin", key, idex);
	fid = PL_open_foreign_frame();
	_do_slave_work(source, &wc1, &wc2, &ru1, &ru2);
	PL_discard_foreign_frame(fid);
	MPI_TRACE("%10s %10lu %10u", "close", key, idex);

	/* Following this is all tracing and recording. */

	timeradd(&ru2.ru_utime, &ru2.ru_stime, &tv2);
	timeradd(&ru1.ru_utime, &ru1.ru_stime, &tv1);

	timersub(&tv2, &tv1, &tv);
	timersub(&wc2, &wc1, &wc);

	/* This can be below zero if the interval is really small. */
	if (wc.tv_sec < 0)
		wc.tv_sec = wc.tv_usec = 0;

	/* These may be less than zero on some systems. */
	if (tv.tv_sec < 0)
		tv.tv_sec = tv.tv_usec = 0;
	tv.tv_usec = tv.tv_usec < 0 ? 0 : tv.tv_usec;
	if (wc.tv_sec < 0)
		wc.tv_sec = wc.tv_usec = 0;
	wc.tv_usec = wc.tv_usec < 0 ? 0 : wc.tv_usec;

	/* Cast aggrs. to larger-width type. */
	tt1 = ((unsigned long long)tv.tv_sec * 1000000) + 
		(unsigned long long)tv.tv_usec;
	tt2 = ((unsigned long long)wc.tv_sec * 1000000) + 
		(unsigned long long)wc.tv_usec;

	if (opts.dump != NULL) {
		if (fprintf(opts.dump, 
				"%10lu %7u %20llu %20llu %10ld "
				"%10ld %10ld %10ld %10ld %10ld\n", 
				key, idex, tt1, tt2,
				ru2.ru_maxrss - ru1.ru_maxrss,
				ru2.ru_maxrss,
				ru2.ru_idrss - ru1.ru_idrss,
				ru2.ru_idrss,
				ru2.ru_isrss - ru1.ru_isrss,
				ru2.ru_isrss) == -1)
			MPI_ABORT();
		if (fflush(opts.dump) == -1)
			MPI_ABORT();
	}
}

void
_do_slave_work(int source, struct timeval *wc1, 
		struct timeval *wc2, struct rusage *ru1, 
		struct rusage *ru2)
{
	unsigned int len;
	predicate_t pred;
	term_t term;
	char *predname, code, *arg, *ep, *out;
	MPI_Request rq;
	long larg;
	double farg;
	int yesno, type;
	assert(wc1 != NULL);
	assert(wc2 != NULL);
	assert(ru1 != NULL);
	assert(ru2 != NULL);

	/* Read in predicate name. */
	MPI_RECV(&len, 1, MPI_UNSIGNED, source, TAG_WORKER_PREDSIZE);
	predname = (char *)calloc(len + 1, 1);
	assert(predname != NULL);
	MPI_RECV(predname, len, MPI_CHAR, source, TAG_WORKER_PRED);

	/* Read in argument type. */
	MPI_RECV(&type, 1, MPI_INT, source, TAG_WORKER_DATATYPE);

	/* Read in arguments. */
	MPI_RECV(&len, 1, MPI_UNSIGNED, source, TAG_WORKER_DATASIZE);
	arg = (char *)calloc(len + 1, 1);
	assert(arg != NULL);
	MPI_RECV(arg, len, MPI_CHAR, source,TAG_WORKER_DATA);

	/* Construct the INPUT predicate terms. */
	pred = PL_predicate(predname, 2, "user");
	term = PL_new_term_refs(2);
	switch(type) {
	case(PL_ATOM):
		PL_put_atom_nchars(term, len, arg);
		break;
	case(PL_STRING):
		PL_put_string_nchars(term, len, arg);
		break;
	case(PL_INTEGER):
		larg = strtol(arg, &ep, 10);
		if (arg[0] == '\0' || *ep != '\0')
			MPI_ABORT();
		if (larg == LONG_MAX || larg == LONG_MIN)
			MPI_ABORT();
		PL_put_integer(term, larg);
		break;
	case(PL_FLOAT):
		farg = strtod(arg, &ep);
		if (arg[0] == '\0' || *ep != '\0')
			MPI_ABORT();
		if (farg == HUGE_VAL || farg == -HUGE_VAL)
			MPI_ABORT();
		PL_put_float(term, farg);
		break;
	default:
		MPI_ABORT();
		/* NOTREACHED */
	}

	/* Actually call the underlying predicate. */
	MPI_DEBUG("+%s(%s)", predname, arg);
	yesno = TRUE;

	/* First time our pre-process profile image. */
	if (gettimeofday(wc1, NULL) == -1)
		MPI_ABORT();
	if (getrusage(RUSAGE_SELF, ru1) == -1)
		MPI_ABORT();

	if ( ! PL_call_predicate(NULL, PL_Q_NORMAL, pred, term))
		yesno = FALSE;

	/* Now post-process profile image. */
	if (getrusage(RUSAGE_SELF, ru2) == -1)
		MPI_ABORT();
	if (gettimeofday(wc2, NULL) == -1)
		MPI_ABORT();

	MPI_DEBUG("-%s(%s) (%d)", predname, arg, yesno);

	/* Free our allocations. */
	free(predname);
	free(arg);

	/* Send result trigger. */
	code = TAG_EVENT_COMPLETE;
	MPI_ISEND(&code, 1, MPI_CHAR, RANK_CONTROL, TAG_EVENT, &rq);
	MPI_ISEND(&yesno, 1, MPI_INT, RANK_CONTROL, 
			TAG_COMPLETE_YESNO, &rq);

	/* Return immediately (no result) if the predicate failed. */
	if ( ! yesno)
		return;

	type = PL_term_type(term + 1);
	switch(type) {
	case(PL_ATOM):
		/* FALLTHROUGH */
	case(PL_INTEGER):
		/* FALLTHROUGH */
	case(PL_FLOAT):
		if ( ! PL_get_chars(term + 1, &out, 
					CVT_ALL | REP_UTF8))
			MPI_ABORT();
		len = strlen(out);
		break;
	case(PL_STRING):
		if ( ! PL_get_string_chars(term + 1, &out, &len))
			MPI_ABORT();
		break;
	case(PL_TERM):
		/* FALLTHROUGH */
	case(PL_VARIABLE):
		MPI_WARN("unknown input type %d (known: atom, string, "
				"integer, float)", type);
		MPI_ABORT();
		/* NOTREACHED */
	}
	MPI_ISEND(&type, 1, MPI_INT, RANK_CONTROL, 
			TAG_COMPLETE_RESULTTYPE, &rq);
	MPI_ISEND(&len, 1, MPI_UNSIGNED, RANK_CONTROL, 
			TAG_COMPLETE_RESULTSIZE, &rq);
	MPI_ISEND(out, len, MPI_CHAR, RANK_CONTROL, 
			TAG_COMPLETE_RESULT, &rq);
}

/*extern*/ int
loop_slave(void)
{
	int rc;
	char code;
	MPI_Status st;

	for ( ; ; ) { /* Infinite loop. */
		MPI_RECV_ST(&code, 1, MPI_CHAR, MPI_ANY_SOURCE, 
				TAG_EVENT, &st);
		switch(code) {
		case(TAG_EVENT_HALT):
			goto out;
		case(TAG_EVENT_WORKER):
			do_slave_work(st.MPI_SOURCE);
			break;
		case(TAG_EVENT_LOAD):
			do_slave_load(st.MPI_SOURCE);
			break;
		default:
			MPI_ABORT();
		}
	} /* Infinite loop. */

out: /* Egress point. */
	MPI_RECV(&rc, 1, MPI_INT, RANK_ROOT, TAG_HALT_CODE);
	PL_cleanup(rc);
	MPI_BARRIER();
	MPI_FINALIZE();

	return(rc);
}

