/* $Id: test.c,v 1.36 2007-01-24 12:16:10 kristaps Exp $ */
/*
Copyright (c) 2006 Kristaps Dzonsons
Latvijas Universitaates Matemaatikas un Informaatikas Instituuts
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <limits.h>
#include <mpi.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include <unistd.h>
#include <SWI-Prolog.h>

#include "pl_mpi.h"


/*-[macros]----------------------------------------------------------*/

#define TYPE_STRING			1
#define TYPE_INT			2
#define TYPE_FLOAT			3

#define SIMPLE_TEST			"simpletest"
#define SIMPLE_TEST_FLOAT		"simpletestfloat"
#define SIMPLE_TEST_INT			"simpletestint"
#define WAIT_SIMPLE_TEST		"waitsimpletest"
#define WAIT_SIMPLE_TEST_INT		"waitsimpletestint"
#define FWAIT_SIMPLE_TEST		"fullwaitsimpletest"
#define FWAIT_SIMPLE_TEST_INT		"fullwaitsimpletestint"
#define YESNO_TEST			"yesnotest"
#define YESNO_TEST_INT			"yesnotestint"
#define YESNO_TEST_FLOAT		"yesnotestfloat"
#define WAIT_YESNO_TEST			"waityesnotest"
#define WAIT_YESNO_TEST_INT		"waityesnotestint"
#define BLOCK_SIMPLE_TEST		"blocksimpletest"

/** Size of chunks in predicate input array. */
#define MAX_ATOMSIZE			32
/** Maximum wait time. */
#define MAX_WAITTEST 			20
/** Maximum elements in input list. */
#define MAX_LISTSIZE			4096
/** Maximum elements in asynchronous query. */
#define MAX_ASYNCSIZE			10

#define PRINTF(fmt, ...) do { \
	(void)printf(fmt, ##__VA_ARGS__); \
	(void)fflush(stdout); } while(0)

/*-[types]-----------------------------------------------------------*/

union input_atom {
	int integer;
	double number;
	char string[MAX_ATOMSIZE];
};

#define	FAST				1
#define	BURN				2
#define	WAIT				4

/*-[variable definitions]--------------------------------------------*/

/*extern*/ unsigned int rank = 0;
/*extern*/ unsigned int nodes = 0;
/*extern*/ struct options opts;

#define RANDOM_POOL_SZ		4096
/** Use a pool of random values, so we needn't be constantly drawing on
 * random(4), which appears to cause issues if over-used.  Since our
 * randomness isn't *very* important, this is acceptable. */
static long random_pool[RANDOM_POOL_SZ];

/*-[function definitions]--------------------------------------------*/

static long
get_rand(void)
{
	struct timeval tm;
	int rc;
	/* This returns a suitably random value. */
	rc = gettimeofday(&tm, NULL);
	assert(rc != -1);
	return(random_pool[tm.tv_usec % (RANDOM_POOL_SZ - 1)]);
}

static void
init_rand(void)
{
	struct timeval tm;
	int rc, i;
	/* This returns a suitably random value. */
	rc = gettimeofday(&tm, NULL);
	assert(rc != -1);
	srandom((unsigned int)tm.tv_usec);
	/* Fill our random pool. */
	for (i = 0; i < RANDOM_POOL_SZ; i++)
		random_pool[i] = random();
}

static int
chars2int(const char *input)
{
	long lval;
	char *ep;
	assert(input != NULL);
	/* From OpenBSD's strtol(3). */
	lval = strtol(input, &ep, 10);
	if (input[0] == '\0' || *ep != '\0')
		MPI_ABORT();
	if ((errno == ERANGE && (lval == LONG_MAX || 
					lval == LONG_MIN)) ||
			(lval > INT_MAX || lval < INT_MIN))
		MPI_ABORT();
	return((int)lval);
}

static int
termchars2int(term_t in)
{
	char *input;
	if ( ! PL_get_atom_chars(in, &input))
		MPI_ABORT();
	return(chars2int(input));
}

static void
burn(long waitt)
{
	time_t tm, t;
	register int i, j;
	tm = time(NULL);
	assert(tm != (time_t)-1);
	assert(waitt >= 0 && waitt < INT_MAX);
	tm += waitt;
	for (;;) {
		t = time(NULL);
		assert(t != (time_t)-1);
		if (t >= tm)
			break;
		for (i = 0; i < 20000; i++)
			for (j = 0; j < 20000; j++)
				(void)(i * j);
	}
}

/*anon*/ foreign_t
yesnotestint(term_t in, term_t out)
{
	int node;
	MPI_DEBUG("%s", __func__);
	if (PL_get_integer(in, &node) == 0)
		MPI_ABORT();
	if (node % 2 == 0)
		PL_fail;
	if ( ! PL_unify_integer(out, 0))
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t
yesnotestfloat(term_t in, term_t out)
{
	double node;
	MPI_DEBUG("%s", __func__);
	if (PL_get_float(in, &node) == FALSE)
		MPI_ABORT();
	if ((int)node % 2 == 0)
		PL_fail;
	if ( ! PL_unify_float(out, 0))
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t
yesnotest(term_t in, term_t out)
{
	int node;
	MPI_DEBUG("%s", __func__);
	node = termchars2int(in);
	if (node % 2 == 0)
		PL_fail;
	if ( ! PL_unify_string_nchars(out, 1, "0"))
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t
waityesnotestint(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (yesnotestint(in, out) == FALSE)
		PL_fail;
	burn(get_rand() % MAX_WAITTEST);
	PL_succeed;
}

/*anon*/ foreign_t
waityesnotest(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (yesnotest(in, out) == FALSE)
		PL_fail;
	burn(get_rand() % MAX_WAITTEST);
	PL_succeed;
}

/*anon*/ foreign_t 
simpletestint(term_t in, term_t out)
{
	int val;
	MPI_DEBUG("%s", __func__);
	if (PL_get_integer(in, &val) == 0)
		MPI_ABORT();
	if (PL_unify_integer(out, val) == 0)
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t 
simpletestfloat(term_t in, term_t out)
{
	double val;
	MPI_DEBUG("%s", __func__);
	if (PL_get_float(in, &val) == FALSE)
		MPI_ABORT();
	if (PL_unify_float(out, val) == 0)
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t 
simpletest(term_t in, term_t out)
{
	int val, sz;
	static char buf[32];
	MPI_DEBUG("%s", __func__);
	val = termchars2int(in);
	sz = snprintf(buf, 32, "%d", val);
	assert(sz >= 0);
	if (PL_unify_string_nchars(out, (unsigned int)sz, buf) == 0)
		MPI_ABORT();
	PL_succeed;
}

/*anon*/ foreign_t 
fullwaitsimpletestint(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (simpletestint(in, out) == FALSE)
		PL_fail;
	burn(MAX_WAITTEST);
	PL_succeed;
}

/*anon*/ foreign_t 
fullwaitsimpletest(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (simpletest(in, out) == FALSE)
		PL_fail;
	burn(MAX_WAITTEST);
	PL_succeed;
}

/*anon*/ foreign_t 
waitsimpletestint(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (simpletestint(in, out) == FALSE)
		PL_fail;
	burn(get_rand() % MAX_WAITTEST);
	PL_succeed;
}

/*anon*/ foreign_t 
waitsimpletest(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (simpletest(in, out) == FALSE)
		PL_fail;
	burn(get_rand() % MAX_WAITTEST);
	PL_succeed;
}

/*anon*/ foreign_t 
blocksimpletest(term_t in, term_t out)
{
	MPI_DEBUG("%s", __func__);
	if (simpletest(in, out) == FALSE)
		PL_fail;
	assert(MAX_WAITTEST < UINT_MAX);
	(void)sleep(get_rand() % MAX_WAITTEST);
	PL_succeed;
}

/*-------------------------------------------------------------------*/

static int
do_eval_yesnotest(const char *name, int res, int yesno, int node)
{
	(void)res;
	if (yesno == ((node % 2 == 0) ? FALSE : TRUE))
		return(TRUE);
	(void)fprintf(stderr, "Error: %d != %d (%s).\n", yesno, 
			(node % 2 == 0 ? FALSE : TRUE), name);
	(void)fflush(stderr);
	return(FALSE);
}

static int
do_eval_simpletest(const char *name, int res, int yesno, int node)
{
	if (yesno != TRUE)
		return(FALSE);
	if (res == node) 
		return(TRUE);
	(void)fprintf(stderr, "Error: %d != %d (%s).\n", 
			res, node, name);
	(void)fflush(stderr);
	return(FALSE);
}

static void
put_list(term_t l, int argc, const union input_atom argv[MAX_LISTSIZE], 
		unsigned int input_type)
{ 
	term_t a;
	a = PL_new_term_ref();
	PL_put_nil(l);
	while( --argc >= 0 ) {
		switch(input_type) {
		case(TYPE_STRING):
			PL_put_atom_chars(a, argv[argc].string);
			break;
		case(TYPE_INT):
			PL_put_integer(a, argv[argc].integer);
			break;
		case(TYPE_FLOAT):
			PL_put_float(a, argv[argc].number);
			break;
		default:
			MPI_ABORT();
			/* NOTREACHED */
		}
		PL_cons_list(l, a, l);
	}
}

/** Fetch aget/dist output values.  Note that "term" must be set to the
 * first output (e.g., term + 1 for aget). */
static void
switch_output(term_t term, unsigned int input_type,
		int *yesno, int *idex, int *result)
{
	char *args;
	double argd;
	unsigned int argssz;
	assert(yesno != NULL);
	assert(idex != NULL);
	assert(result != NULL);

	if ( ! PL_get_integer(term + 1, yesno))
		MPI_ABORT();
	if ( ! PL_get_integer(term + 2, idex))
		MPI_ABORT();

	/* No solution from the index.  Return now. */
	if ( ! *yesno)
		return;

	/* Convert output data to an integer. */
	switch(input_type) {
	case(TYPE_STRING):
		if ( ! PL_get_string_chars(term, &args, &argssz))
			MPI_ABORT();
		*result = chars2int(args);
		break;
	case(TYPE_INT):
		if ( ! PL_get_integer(term, result))
			MPI_ABORT();
		break;
	case(TYPE_FLOAT):
		if ( ! PL_get_float(term, &argd))
			MPI_ABORT();
		*result = (int)argd;
		break;
	default:
		MPI_ABORT();
	}
}

/** Fill a single input atom by converting to its type. */
static void
switch_input(unsigned int input_type, union input_atom *res,
		int value)
{
	int rc;
	assert(res != NULL);
	assert(value >= 0);
	switch(input_type) {
	case(TYPE_STRING):
		rc = snprintf(res->string, MAX_ATOMSIZE, "%d", value);
		if (rc <= 0 || rc >= MAX_ATOMSIZE - 1)
			MPI_ABORT();
		break;
	case(TYPE_INT):
		res->integer = value;
		break;
	case(TYPE_FLOAT):
		res->number = (double)value;
		break;
	default:
		MPI_ABORT();
	}
}

/** Initialise the asynchronous queries. */
static void
do_adist(union input_atom input[MAX_ASYNCSIZE][MAX_LISTSIZE],
		size_t size, size_t async, const char *predname,
		unsigned int input_type,
		int64_t output[MAX_ASYNCSIZE])
{
	predicate_t pred;
	term_t term;
	functor_t func;
	fid_t fid;
	size_t i;
	assert(predname != NULL);

	fid = PL_open_foreign_frame();

	/* Prepare the term sequence. */
	pred = PL_predicate(PREDICATE_ADIST, 3, "user");
	func = PL_new_functor(PL_new_atom(predname), 2);

	/* Concurrently send out the asynchronous queries. */
	assert(async < INT_MAX);
	for (i = 0; i < async; i++) {
		term = PL_new_term_refs(3);
		PL_put_functor(term, func);
		put_list(term + 1, size, input[i], input_type);
		/* Call for a single solution.  This should never return
		 * the FALSE token. */
		if ( ! PL_call_predicate(NULL, PL_Q_NORMAL, pred, term))
			MPI_ABORT();
		/* Fill the array of qids. */
		if ( ! PL_get_int64(term + 2, &output[i]))
			MPI_ABORT();
	}

	PL_discard_foreign_frame(fid);
}

/** Close pending asynchronous queries. */
static void
do_aclose(int64_t qid[MAX_ASYNCSIZE], size_t count)
{
	size_t i;
	predicate_t pred;
	term_t term;
	fid_t fid;

	fid = PL_open_foreign_frame();

	/* Prepare the term sequence. */
	pred = PL_predicate(PREDICATE_ACLOSE, 1, "user");
	term = PL_new_term_refs(1);

	for (i = 0; i < count; i++) {
		PL_put_int64(term, qid[i]);
		/* Call the predicate.  Return FALSE immediately if the
		 * predicate returns no solution. */
		if ( ! PL_call_predicate(NULL, PL_Q_NORMAL, pred, term))
			MPI_ABORT();
	}
	PL_discard_foreign_frame(fid);
}

/** Fetch pending results of an asynchronous query. 
 * @return TRUE if the predicate returns TRUE else FALSE. */
static int
do_aget(int64_t qid, unsigned int input_type, int *yesno, 
		int *idex, int *res)
{
	predicate_t pred;
	term_t term;
	fid_t fid;
	assert(yesno != NULL);
	assert(idex != NULL);
	assert(res != NULL);

	fid = PL_open_foreign_frame();

	/* Prepare the term sequence. */
	pred = PL_predicate(PREDICATE_AGET, 4, "user");
	term = PL_new_term_refs(4);
	PL_put_int64(term + 0, qid);

	/* Call the predicate.  Return FALSE immediately if the
	 * predicate returns no solution. */
	if ( ! PL_call_predicate(NULL, PL_Q_NORMAL, pred, term))
		return(0);
	switch_output(term + 1, input_type, yesno, idex, res);
	PL_discard_foreign_frame(fid);
	return(1);
}

/** @return TRUE on success, FALSE on failure. */
static int
do_fixture_sync(unsigned long num, const char *predname, 
		unsigned int input_type, size_t size, size_t limit,
		int(*eval)(const char *, int, int, int))
{
	int i, res, idex, yesno, recvd;
	static union input_atom input[MAX_LISTSIZE];
	static int res_array[MAX_LISTSIZE];
	term_t term;
	qid_t query;
	predicate_t pred;
	functor_t func;

	PRINTF("\033[0G\033[K%5lu %-26s %6zd, %-8zd %11s", 
			num, predname, size, limit, "[init]");

	/* Fill the input array with our test values. */
	assert(size < INT_MAX);
	for (i = 0; i < (int)size; i++) 
		switch_input(input_type, &input[i], i);

	/* Prepare the term sequence. */
	pred = PL_predicate(PREDICATE_DIST, 5, "user");
	term = PL_new_term_refs(5);
	assert(predname != NULL);
	func = PL_new_functor(PL_new_atom(predname), 2);
	PL_put_functor(term + 0, func);
	put_list(term + 1, size, input, input_type);
	(void)memset(res_array, 0, sizeof(res_array));

	/* Construct our query! */
	query = PL_open_query(NULL, PL_Q_NORMAL, pred, term);

	PRINTF("\033[0G\033[K%5lu %-26s %6zd, %-8zd %10.2f%%",
			num, predname, size, limit, 0.0);

	assert(limit <= size);
	for (recvd = 0, i = 0; i < (int)limit && 
			PL_next_solution(query) != FALSE; i++) {
		switch_output(term + 2, input_type, 
				&yesno, &idex, &res);
		/* In this case, we've already received a response for
		 * this node.  Duplicates are bad. */
		if (res_array[idex]) {
			PRINTF(" [index %d duplicate]", idex);
			goto fail;
		}
		res_array[idex] = 1;
		recvd++;

		/* Note that "res" may be unset if there is no solution
		 * for this query index. */
		if ((*eval)(predname, res, yesno, idex) == FALSE)
			goto fail;

		/* Success for this test entry. */
		PRINTF("\033[0G\033[K%5lu %-26s %6zd, "
				"%-8zd %10.2f%%", 
				num, predname, size, limit, 
				((double)i / limit) * 100);
	}
	/* (Remember to close out query!) */
	PL_close_query(query);

	/* TODO: make sure we haven't received more than the number of
	 * available results. */

	if (recvd < (int)limit) {
		PRINTF(" [missing %zd results]", limit - recvd);
		goto fail;
	} 
	PRINTF("\033[0G\033[K%5lu %-26s %6zd, %-8zd %11s\n",
			num, predname, size, limit, "[success]");
	return(TRUE);
	
fail:	/* Failure clause. */
	PRINTF("\n");
	return(FALSE);
}

/** @return TRUE on success, FALSE on failure. */
static int
do_fixture_async(unsigned long num, const char *predname, 
		unsigned int input_type, size_t size, size_t limit,
		int(*eval)(const char *, int, int, int),
		size_t async)
{
	size_t j, i, k;
	static union input_atom input[MAX_ASYNCSIZE][MAX_LISTSIZE];
	int64_t qids[MAX_ASYNCSIZE];
	int yesno, idex, res;

	PRINTF("\033[0G\033[K%5lu %-21s %4zd %6zd, %-8zd %11s",
			num, predname, async, size, limit, "[init]");

	/* TODO: variegate array sizes. */
	/* Fill the input array with our test value. */
		for (j = 0; j < async; j++) 
		for (i = 0; i < size; i++) 
			switch_input(input_type, &input[j][i], i);

	do_adist(input, size, async, predname, input_type, qids);

	PRINTF("\033[0G\033[K%5lu %-21s %4zd %6zd, %-8zd %10.2f%%",
			num, predname, async, size, limit, 0.0);

	for (k = 0, i = 0; i < limit; i++) {
		for (j = 0; j < async; j++, k++) {
			/* TODO: border-check that checks after "async"
			 * tries execute PL_fail. */
			/* TODO: have a "res_array" and "recvd" like in
			 * the synchronous version. */
			if ( ! do_aget(qids[j], input_type, &yesno,
						&idex, &res)) {
				goto fail;
			}
			if ((*eval)(predname, res, yesno, idex) 
					== FALSE)
				goto fail;
			PRINTF("\033[0G\033[K%5lu %-21s %4zd "
					"%6zd, %-8zd %10.2f%%", 
				num, predname, async, size, limit, 
				((double)k / (limit + async)) 
					* 100);
		}
	}

	PRINTF("\033[0G\033[K%5lu %-21s %4zd %6zd, "
			"%-8zd %11s", num, predname, 
			async, size, limit, "[cleanup]");
	do_aclose(qids, async);
	PRINTF("\033[0G\033[K%5lu %-21s %4zd %6zd, "
			"%-8zd %11s\n", num, predname, 
			async, size, limit, "[success]");
	return(TRUE);
fail:

	PRINTF("\n");
	return(FALSE);
}

/** Routes to the appropriately fixture subroutine.
 * @return TRUE on success, FALSE on failure. */
static int
do_fixture(unsigned long num, const char *predname, 
		unsigned int input_type, size_t size, size_t limit,
		int(*eval)(const char *, int, int, int),
		unsigned int async)
{
	return(async == 0 ?
		do_fixture_sync(num, predname, input_type,
			size, limit, eval) :
		do_fixture_async(num, predname, input_type,
			size, limit, eval, async));
}

static int
do_test(int uniform, size_t size, int prof, 
		unsigned long count, unsigned int asyncsize)
{
	/** Number of elements in predicate/data array. */
#define TARGET_SIZE 11
	struct target_entry {
		int type;
		/**< Type of input/output (TYPE_STRING, ...). */
		const char *predicate;
		/**< Name of predicate. */
		int (*fp)(const char *, int, int, int);
		/**< Output evaluation function. */
	} targets[TARGET_SIZE] = {
		{ TYPE_STRING, SIMPLE_TEST, do_eval_simpletest },
		{ TYPE_FLOAT, SIMPLE_TEST_FLOAT, do_eval_simpletest },
		{ TYPE_INT, SIMPLE_TEST_INT, do_eval_simpletest },
		{ TYPE_STRING, YESNO_TEST, do_eval_yesnotest },
		{ TYPE_FLOAT, YESNO_TEST_FLOAT, do_eval_yesnotest },
		{ TYPE_INT, YESNO_TEST_INT, do_eval_yesnotest },
		{ TYPE_STRING, WAIT_SIMPLE_TEST, do_eval_simpletest },
		{ TYPE_INT, WAIT_SIMPLE_TEST_INT, do_eval_simpletest },
		{ TYPE_STRING, FWAIT_SIMPLE_TEST, do_eval_simpletest },
		{ TYPE_INT, FWAIT_SIMPLE_TEST_INT, do_eval_simpletest },
		{ TYPE_STRING, BLOCK_SIMPLE_TEST, do_eval_simpletest },
	}, *e;
	size_t retr, localsize;
		/* retr = number to retrieve from localsize */
		/* localsize = input elements (random from set domain) */
	unsigned long test;
	unsigned int async, doasync;

	(void)printf("%6s %-21s %4s %6s, %s %15s\n", 
			" ", "predicate", "sync", "size", "retr", "status");

	/* "Infinite" tests are really ULONG_MAX. */
	count = count == 0 ? ULONG_MAX : count;
	for (test = 0; test < count; test++) {
		localsize = (size_t)(get_rand() % size);
		/* Normalise to be non-zero. */
		localsize = localsize == 0 ? 1 : localsize;
		retr = uniform ? localsize : (get_rand() % localsize);
		/* Normalise to be non-zero. */
		retr = retr == 0 ? 1 : retr;
		/* See if we're going to be asynchronous. */
		doasync = get_rand() % 2;
		async = doasync ? (get_rand() % asyncsize) : 0;

		/* Choose from test profile. */
		if (prof == 0) 
			e = &targets[get_rand() % TARGET_SIZE];
		else {
			int domain[TARGET_SIZE], i, offs = 0;
			/* Establish a mapping to our problem domain.
			 * These values will change if the target array
			 * changes, so be warned. */
			if (prof & FAST) 
				for (i = 0; i < 6; i++, offs++)
					domain[offs] = i;
			if (prof & BURN)
				for (i = 6; i < 10; i++, offs++)
					domain[offs] = i;
			if (prof & WAIT)
				domain[offs++] = 10;
			offs--;
			e = &targets[domain[get_rand() & offs]];
		}
		if ( ! do_fixture(test, e->predicate, e->type, 
					localsize, retr, e->fp, async))
			return(-1);
	}
	return(0);
}

static void
usage(int code)
{
	extern char *__progname;
	if (rank == 0)
		(void)fprintf(stderr, "Usage: %s "
				"[-dut] "
				"[-a asyncsize] "
				"[-p profile] "
				"[-r num] "
				"[-s size]\n", __progname);
	MPI_FINALIZE();
	exit(code);
	/* NOTREACHED */
}

static int
parse_profile(char *optarg)
{
	char *p;
	int prof;
	for (p = optarg, prof = 0; ; ) {
		/* Check for four-letter profiles. */
		if (strlen(optarg) < 4)
			return(-1);
		if (strncasecmp(p, "fast", 4) == 0)
			prof |= FAST;
		else if (strncasecmp(p, "burn", 4) == 0)
			prof |= BURN;
		else if (strncasecmp(p, "wait", 4) == 0)
			prof |= WAIT;
		else 
			return(-1);
		p += 4;
		if (*p == '\0') 
			break;
		if (*p++ != ',')
			return(-1);
	}
	return(prof);
}


int
main(int argc, char *argv[])
{ 
	PL_extension pt[] = {
		{ SIMPLE_TEST, 2, simpletest, 0 },
		{ SIMPLE_TEST_FLOAT, 2, simpletestfloat, 0 },
		{ SIMPLE_TEST_INT, 2, simpletestint, 0 },
		{ WAIT_SIMPLE_TEST, 2, waitsimpletest, 0 },
		{ WAIT_SIMPLE_TEST_INT, 2, waitsimpletestint, 0 },
		{ YESNO_TEST, 2, yesnotest, 0 },
		{ YESNO_TEST_INT, 2, yesnotestint, 0 },
		{ YESNO_TEST_FLOAT, 2, yesnotestfloat, 0 },
		{ WAIT_YESNO_TEST, 2, waityesnotest, 0 },
		{ WAIT_YESNO_TEST_INT, 2, waityesnotestint, 0 },
		{ FWAIT_SIMPLE_TEST, 2, fullwaitsimpletest, 0 },
		{ FWAIT_SIMPLE_TEST_INT, 2, fullwaitsimpletestint, 0 },
		{ BLOCK_SIMPLE_TEST, 2, blocksimpletest, 0 },
		{ NULL, 0, NULL, 0 }
	};
	extern int optind;
	int c, rc, uniform, speed;
	size_t size;
	unsigned long randrun;
	unsigned int asyncsize;
	const char *errstr; /* for strtonum */

	(void)memset(&opts, 0, sizeof(struct options));

	/* Subsystem initialisations. */
	init_mpi(&argc, &argv);
	init_prolog(argc, argv);
	init_rand();

	PL_register_extensions_in_module("user", pt);

	(void)PL_signal(PL_SIGSYNC | SIGSEGV, (void(*)(int))abort);

	if (rank == 0)
		setproctitle("master");
	else if (rank == 1)
		setproctitle("control");
	else
		setproctitle("slave [rank %d]", rank);

	if (snprintf(opts.dump_file, MAXPATHLEN, "dump-%d.txt", 
				rank) >= MAXPATHLEN - 1)
		errx(1, "profile filename too long");
	if (snprintf(opts.trace_file, MAXPATHLEN, "trace-%u.txt",
				rank) >= MAXPATHLEN - 1)
		errx(1, "trace filename too long");

	/* Option parsing. */
	randrun = 0;
	size = MAX_LISTSIZE;
	asyncsize = MAX_ASYNCSIZE;
	speed = 0;
	uniform = 0;
	errstr = NULL;
	while ((c = getopt(argc, argv, "us:r:tp:da:")) != -1) 
		switch (c) {
		case('p'):
			if ((speed = parse_profile(optarg)) == -1)
				errx(1, "invalid profile: -p %s",
						optarg);
			break;
		case('a'):
			asyncsize = (unsigned int)strtonum
				(optarg, 1, UINT_MAX, &errstr);
			if (errstr)
				err(1, "max asynchronous (-a): %s, %s",
						errstr, optarg);
			break;
		case('s'):
			size = (size_t)strtonum
				(optarg, 1, INT_MAX, &errstr);
			if (errstr)
				err(1, "test width (-s): %s, %s",
						errstr, optarg);
			break;
		case('d'):
			opts.dump = fopen(opts.dump_file, "w");
			if (opts.dump == NULL)
				err(1, "failed open dump: %s",
						opts.dump_file);
			break;
		case('t'):
			opts.trace = fopen(opts.trace_file, "w");
			if (opts.trace == NULL)
				err(1, "failed open trace: %s",
						opts.trace_file);
			break;
		case('u'):
			uniform = 1;
			break;
		case('r'):
			randrun = (unsigned long)strtonum
				(optarg, 0, LONG_MAX, &errstr);
			if (errstr)
				err(1, "run count(-r): %s, %s",
						errstr, optarg);
			break;
		default:
			usage(1);
			/* NOTREACHED */
		}
	argc -= optind;
	argv += optind;

	MPI_BARRIER();

	/* Route into the main caller. */
	switch(rank) {
	case(0):
		PL_halt(do_test(uniform, size, speed, 
					randrun, asyncsize));
		/* NOTREACHED */
	case(1):
		rc = loop_cntrl();
		break;
	default:
		rc = loop_slave();
		break;
	}

	return(rc);
}


